from collections import defaultdict
import re, sched, time
from typing import Any, Callable, Dict, List, Optional, Tuple
from croniter import croniter
import numpy as np
import pandas as pd
from threading import Thread

from qubx import logger
from qubx.core.basics import CtrlChannel
from qubx.utils.misc import Stopwatch
from qubx.utils.time import convert_tf_str_td64, convert_seconds_to_str
from qubx.core.series import TimeSeries, Trade, Quote, Bar, OHLCV


_SW = Stopwatch()

class CachedMarketDataHolder: 
    """
    Collected cached data updates from StrategyContext
    """
    default_timeframe: np.timedelta64
    _last_bar: Dict[str, Bar | None]
    _ohlcvs: Dict[str, Dict[np.timedelta64, OHLCV]]
    _updates: Dict[str, Any]

    def __init__(self, default_timeframe: str) -> None:
        self.default_timeframe = convert_tf_str_td64(default_timeframe)
        self._ohlcvs = dict()
        self._last_bar = defaultdict(lambda: None)
        self._updates = dict()

    def init_ohlcv(self, symbol: str, max_size=np.inf):
        self._ohlcvs[symbol] = {self.default_timeframe: OHLCV(symbol, self.default_timeframe, max_size)}

    def is_data_ready(self) -> bool:
        """
        Check if all symbols in this cache have at least one update
        """
        for v in self._ohlcvs.keys():
            if v not in self._updates:
                return False
        return True
    
    @_SW.watch('CachedMarketDataHolder')
    def get_ohlcv(self, symbol: str, timeframe: str, max_size=np.inf) -> OHLCV:
        tf = convert_tf_str_td64(timeframe) 

        if symbol not in self._ohlcvs:
           self._ohlcvs[symbol] = {}

        if tf not in self._ohlcvs[symbol]: 
            # - check requested timeframe
            new_ohlc = OHLCV(symbol, tf, max_size)
            if tf < self.default_timeframe:
                logger.warning(f"[{symbol}] Request for timeframe {timeframe} that is smaller then minimal {self.default_timeframe}")
            else:
                # - first try to resample from smaller frame
                if (basis := self._ohlcvs[symbol].get(self.default_timeframe)):
                    for b in basis[::-1]:
                        new_ohlc.update_by_bar(b.time, b.open, b.high, b.low, b.close, b.volume, b.bought_volume)
                
            self._ohlcvs[symbol][tf] = new_ohlc

        return self._ohlcvs[symbol][tf]

    @_SW.watch('CachedMarketDataHolder')
    def update_by_bars(self, symbol: str, timeframe: str, bars: List[Bar]) -> OHLCV:
        """
        Substitute or create new series based on provided historical bars
        """
        if symbol not in self._ohlcvs:
           self._ohlcvs[symbol] = {}

        tf = convert_tf_str_td64(timeframe) 
        new_ohlc = OHLCV(symbol, tf)
        for b in bars:
            new_ohlc.update_by_bar(b.time, b.open, b.high, b.low, b.close, b.volume, b.bought_volume)
            self._updates[symbol] = b

        self._ohlcvs[symbol][tf] = new_ohlc
        return new_ohlc

    @_SW.watch('CachedMarketDataHolder')
    def update_by_bar(self, symbol: str, bar: Bar):
        self._updates[symbol] = bar

        _last_bar = self._last_bar[symbol]
        v_tot_inc = bar.volume
        v_buy_inc = bar.bought_volume

        if _last_bar is not None:
            if _last_bar.time == bar.time: # just current bar updated
                v_tot_inc -= _last_bar.volume
                v_buy_inc -= _last_bar.bought_volume

            if _last_bar.time > bar.time: # update is too late - skip it
                return

        if symbol in self._ohlcvs:
            self._last_bar[symbol] = bar
            for ser in self._ohlcvs[symbol].values():
                ser.update_by_bar(bar.time, bar.open, bar.high, bar.low, bar.close, v_tot_inc, v_buy_inc)

    @_SW.watch('CachedMarketDataHolder')
    def update_by_quote(self, symbol: str, quote: Quote):
        self._updates[symbol] = quote

        series = self._ohlcvs.get(symbol)
        if series:
            for ser in series.values():
                ser.update(quote.time, quote.mid_price(), 0)

    @_SW.watch('CachedMarketDataHolder')
    def update_by_trade(self, symbol: str, trade: Trade):
        self._updates[symbol] = trade
        series = self._ohlcvs.get(symbol)
        if series:
            total_vol = trade.size 
            bought_vol = total_vol if trade.taker >= 1 else 0.0 
            for ser in series.values():
                ser.update(trade.time, trade.price, total_vol, bought_vol)


SPEC_REGEX = re.compile(
    r"((?P<type>[A-Za-z]+)(\.?(?P<timeframe>[0-9A-Za-z]+))?\ *:)?"
    r"\ *"
    r"((?P<spec>"
        r"(?P<time>((\d+:\d+(:\d+)?)\ *,?\ *)+)?"
        r"((\ *@\ *)(?P<by>([A-Za-z0-9-,\ ]+)))?"
        r"(("
        r'((?P<months>[-+]?\d+)(months|month|bm|mo))?'
        r'((?P<weeks>[-+]?\d+)(weeks|week|w))?'
        r'((?P<days>[-+]?\d+)(days|day|d))?'
        r'((?P<hours>[-+]?\d+)(hours|hour|h))?'
        r'((?P<minutes>[-+]?\d+)(mins|min|m))?'
        r'((?P<seconds>[-+]?\d+)(sec|s))?'
        r")(\ *)?)*"
        r".*"
    r"))?", re.IGNORECASE
)


def _mk_cron(time: str, by: list | None) -> str:
    HMS = lambda s: list(map(int, s.split(':') if s.count(':') == 2 else [*s.split(':'), 0]))

    h,m,s = HMS(time)
    assert h < 24, f'Wrong value for hour {h}'
    assert m < 60, f'Wrong value for minute {m}'
    assert s < 60, f'Wrong value for seconds {s}'
    b = ','.join(by) if by else '*'
    c = f'{m} {h} * * {b}'
    return c if s == 0 else c + f' {s}'


def _make_shift(_b, _w, _d, _h, _m, _s):
    D0 = pd.Timedelta(0)
    AS_TD = lambda d: pd.Timedelta(d)
    P, N = D0, D0 
    
    # return AS_TD(f'{_b*4}W') + AS_TD(f'{_w}W') + AS_TD(f'{_d}D') + AS_TD(f'{_h}h') + AS_TD(f'{_m}Min') + AS_TD(f'{_s}Sec')
    for t in [
        AS_TD(f'{_b*4}W'), AS_TD(f'{_w}W'), AS_TD(f'{_d}D'), 
        AS_TD(f'{_h}h'), AS_TD(f'{_m}Min'), AS_TD(f'{_s}Sec')]:
        if t > D0:
            P += t
        else:
            N += t
    return P, N


def _parse_schedule_spec(schedule: str) -> Dict[str, str]:
    m = SPEC_REGEX.match(schedule)
    return {k: v for k, v in m.groupdict().items() if v} if m else {}


def process_schedule_spec(spec_str: str | None) -> Dict[str, Any]:
    AS_INT = lambda d, k: int(d.get(k, 0)) 
    S = lambda s: [x for x in re.split(r"[, ]", s) if x]
    config = {}

    if not spec_str:
        return config

    # - parse schedule spec
    spec = _parse_schedule_spec(spec_str)

    # - check how to run it 
    _T, _S = spec.get('type'), spec.get('spec')
    _F = spec.get('timeframe')
    _t, _by = S(spec.get('time', '')), S(spec.get('by', ''))
    _b, _w, _d = AS_INT(spec, 'months'), AS_INT(spec, 'weeks'), AS_INT(spec, 'days')
    _h, _m, _s = AS_INT(spec, 'hours'), AS_INT(spec, 'minutes'), AS_INT(spec, 'seconds')
    _has_intervals = (_b != 0) or (_w != 0) or (_d != 0) or (_h != 0) or (_m != 0) or (_s != 0) 
    _s_pos, _s_neg = _make_shift(_b, _w, _d, _h, _m, _s)
    _shift = _s_pos + _s_neg

    match _T:
        case 'cron':
            if not _S or croniter.is_valid(_S):
                config = dict(type='cron', schedule=_S, spec=_S)
            else:
                raise ValueError(f"Wrong specification for cron type: {_S}")

        case 'time':
            for t in _t:
                config = dict(type='cron', schedule=_mk_cron(t, _by), spec=_S)
        
        case None:
            if _t: # - if time specified
                for t in _t:
                    config = dict(type='cron', schedule=_mk_cron(t, _by), spec=_S)
            else:
                # - check if it's valid cron
                if _S:
                    if croniter.is_valid(_S):
                        config = dict(type='cron', schedule=_S, spec=_S)
                    else:
                        if _has_intervals:
                            _F = convert_seconds_to_str(int(_s_pos.as_unit('s').to_timedelta64().item().total_seconds())) if not _F else _F  
                            config = dict(type='bar', schedule=None, timeframe=_F, delay=_s_neg, spec=_S)
        case _:
            config = dict(type=_T, schedule=None, timeframe=_F, delay=_shift, spec=_S)

    return config


_SEC2TS = lambda t: pd.Timestamp(t, unit='s')

class BasicScheduler:
    """
    Basic scheduler functionality. It helps to create scheduled event task
    """
    _chan: CtrlChannel
    _scdlr:  sched.scheduler
    _ns_time_fun: Callable[[], float]
    _crons: Dict[str, croniter]
    _is_started: bool

    def __init__(self, channel: CtrlChannel, time_provider_ns: Callable[[], float]):
        self._chan = channel
        self._ns_time_fun = time_provider_ns
        self._scdlr = sched.scheduler(self.time_sec)
        self._crons = dict()
        self._is_started = False

    def time_sec(self) -> float:
        return self._ns_time_fun() / 1000000000.0

    def schedule_event(self, cron_schedule: str, event_name: str):
        if not croniter.is_valid(cron_schedule):
            raise ValueError(f"Specified schedule {cron_schedule} for {event_name} doesn't have valid cron format !")
        self._crons[event_name] = croniter(cron_schedule, self.time_sec())

        if self._is_started:
            self._arm_schedule(event_name, self.time_sec())

    def get_event_last_time(self, event_name: str) -> pd.Timestamp | None:
        if event_name in self._crons: 
            _iter = self._crons[event_name]
            _c = _iter.get_current()
            _t = pd.Timestamp(_iter.get_prev(), unit='s')
            _iter.set_current(_c, force=True)
            return _t
        return None

    def get_event_next_time(self, event_name: str) -> pd.Timestamp | None:
        if event_name in self._crons: 
            _iter = self._crons[event_name]
            _t = pd.Timestamp(_iter.get_next(start_time=self.time_sec()), unit='s')
            return _t
        return None

    def _arm_schedule(self, event: str, start_time: float) -> bool:
        iter = self._crons[event]
        prev_time = iter.get_prev()
        next_time = iter.get_next(start_time=start_time)
        if next_time:
            self._scdlr.enterabs(
                next_time, 1, self._trigger, (event, prev_time, next_time)
            )
            logger.debug(f"Next ({event}) event scheduled at <red>{_SEC2TS(next_time)}</red>")
            return True
        logger.debug(f"({event}) task is not scheduled")
        return False

    def _trigger(self, event: str, prev_time_sec: float, trig_time: float):
        now = self.time_sec()

        # - send notification to channel
        if self._chan.control.is_set():
            self._chan.queue.put((None, event, (prev_time_sec, trig_time)))

        # - try to arm this event again
        self._arm_schedule(event, now)

    def check_and_run_tasks(self) -> float | None:
        return self._scdlr.run(blocking=False) 

    def run(self):
        if self._is_started:
            logger.warning("Scheduler is already running")
            return

        _has_tasks = False
        for k in self._crons.keys():
            _has_tasks |= self._arm_schedule(k, self.time_sec())

        def _watcher():
            while (r := self.check_and_run_tasks()):
                if not self._chan.control.is_set():
                    break
                _delay = max(min(r/5, 5), 0.1)
                time.sleep(_delay)
            logger.debug("Scheduler is stopped ")                    
            self._is_started = False

        if _has_tasks:
            Thread(target=_watcher).start()
            self._is_started = True

    


from typing import Any, Callable, Dict, List, Tuple
from os import unlink
import numpy as np
import pandas as pd
from os.path import exists, join, split, basename
from tqdm.notebook import tqdm
import requests
from collections import defaultdict

from qubx import logger
from qubx.utils.misc import makedirs, get_local_qubx_folder
from qubx.pandaz import generate_equal_date_ranges, srows


DEFALT_LOCAL_FILE_STORAGE = makedirs(get_local_qubx_folder(), 'data/import/binance_history/')
DEFALT_LOCAL_CSV_STORAGE = makedirs(get_local_qubx_folder(), 'data/binance/')

# _DEFAULT_MARKET_DATA_DB = 'md'
BINANCE_DATA_STORAGE = "https://s3-ap-northeast-1.amazonaws.com"
BINANCE_DATA_URL = "https://data.binance.vision/"

    
def get_binance_symbol_info_for_type(market_types: List[str]) -> Dict[str, Dict[str, Any]]:
    """
    Get list of all symbols from binance for given list of market types:
    possible types are: SPOT, FUTURES, COINSFUTURES
    
    >>> get_binance_symbol_info_for_type('FUTURES')
    
    :param market_type: SPOT, FUTURES (UM) or COINSFUTURES (CM)
    """
    from binance.client import Client

    client = Client()
    infos = {}
    for market_type in (market_types if not isinstance(market_types, str) else [market_types]):
        if market_type in ['FUTURES', 'UM']:
            infos['binance.um'] = client.futures_exchange_info()

        elif market_type in ['COINSFUTURES', 'CM']:
            infos['binance.cm'] = client.futures_coin_exchange_info()

        elif market_type == 'SPOT':
            infos['binance'] = client.get_exchange_info()
        else:
            raise ValueError("Only 'FUTURES | UM', 'COINSFUTURES | CM' or 'SPOT' are supported for market_type")

    return infos


def fetch_file(url, local_file_storage, chunk_size=1024*1024, progress_bar=True):
    """
    Load file from url and store it to specified storage
    """
    file = split(url)[-1]
    
    # if dest location not exists create it
    if not exists(local_file_storage):
        makedirs(local_file_storage)
        
    response = requests.get(url, stream=True)
    fpath = join(local_file_storage, file)
    with open(fpath, "wb") as handle:
        iters = response.iter_content(chunk_size=chunk_size)
        for data in tqdm(iters) if progress_bar else iters:
            handle.write(data)
    return fpath


def get_trades_files(symbol: str, instr_type: str, instr_subtype: str):
    """
    Get list of trades files for specified instrument from Binance datastorage
    """
    if instr_type.lower() == 'spot':
        instr_subtype = ''
    filter_str = join("data", instr_type.lower(), instr_subtype.lower(), "monthly", "trades", symbol.upper())
    pg = requests.get(f"{BINANCE_DATA_STORAGE}/data.binance.vision?prefix={filter_str}/")
    info = pd.read_xml(pg.text)
    return [k for k in info.Key.dropna() if k.endswith('.zip')]


def load_trades_for(symbol, instr_type='futures', instr_subtype='um', local_file_storage=DEFALT_LOCAL_FILE_STORAGE):
    """
    Load trades from Binance data storage
    >>> load_trades_for('ETHUSDT', 'futures', 'um')
    """
    local_file_storage = makedirs(local_file_storage)

    f_list = get_trades_files(symbol, instr_type, instr_subtype)
    for r_file in tqdm(f_list):
        dest_dir = join(local_file_storage, symbol.upper())
        dest_file = join(dest_dir, basename(r_file))
        if not exists(dest_file):
            fetch_file(join(BINANCE_DATA_URL, r_file), dest_dir)
        else:
            logger.info(f"{dest_file} already loaded, skipping ...")


def parse_kl_file(fpath: str) -> pd.DataFrame:
    _reader = lambda fp, hdr: pd.read_csv(fp, names = [
            'open_time', 'open', 'high', 'low', 'close', 'volume', 'close_time', 
            'quote_volume', 'count', 'taker_buy_volume', 'taker_buy_quote_volume', 'ignore'
        ], 
        usecols=[
            'open_time', 'open', 'high', 'low', 'close', 'volume', 'quote_volume', 
            'count', 'taker_buy_volume', 'taker_buy_quote_volume', 
        ], 
        index_col='open_time', header=hdr
    )
    d = _reader(fpath, None)

    # - if there is header 
    if isinstance(d.index[0], str): 
        d = d.iloc[1:, :].astype(float)

    d.index = pd.to_datetime(d.index, unit='ms').rename('timestamp')
    return d


def load_binance_kl_history(symbol: str, start: str, stop: str = 'now',
                            instr_type='futures', instr_subtype='um', 
                            timeframe='1m', temp_storage=DEFALT_LOCAL_FILE_STORAGE) -> pd.DataFrame:
    """
    Loads binance 1m KLine history from AWS storage

    Parameters
    ----------
    symbol : str
        The symbol to load data for
    start : str
        The start date in format %Y-%m-%d
    stop : str, optional
        The end date in format %Y-%m-%d, by default 'now'
    instr_type : str, optional
        The instrument type, one of 'futures', 'spot', by default 'futures'
    instr_subtype : str, optional
        The instrument subtype, one of 'um', 'cm', by default 'um'
    timeframe : str, optional
        The timeframe, one of '1m', '3m', '5m', '15m', '30m', '1h', '2h', '4h', '6h', '8h', '12h', '1d', '3d', '1w', '1M', by default '1m'
    temp_storage : str, optional
        The temporary storage location, by default 'DEFALT_LOCAL_FILE_STORAGE'

    Returns
    -------
    pd.DataFrame
        The loaded data
    """
    start = pd.Timestamp(start)
    stop = pd.Timestamp(stop)
    if instr_type.lower() == 'futures':
        subt = f"{instr_type}/{instr_subtype}"
    else:
        subt = 'spot'
    temp_storage =  join(temp_storage, subt)
 
    def _loader(start: pd.Timestamp, stop: pd.Timestamp, f_units: str):
        curr_date = pd.Timestamp('now').ceil('1d')
        data = pd.DataFrame()
        continue_from = None
        for t, _ in generate_equal_date_ranges(start, stop, 1, f_units[0].upper()):
            if f_units =='monthly':
                dt = pd.Timestamp(t)
                # stop when we got into current month
                if dt.year == curr_date.year and dt.month == curr_date.month:
                    continue_from = t
                    break
                dt = dt.strftime("%Y-%m")
            else:
                dt = pd.Timestamp(t).strftime("%Y-%m-%d")

            fname = f"{symbol.upper()}-{timeframe}-{dt}.zip"
            zf = join(temp_storage, fname)

            if not exists(zf):
                logger.info(f'[green]{(symbol)}[/green] {instr_subtype} {instr_type} loading data for [yellow]{dt}[/yellow] -> [red]{fname}[/red]')
                u = f'{BINANCE_DATA_URL}data/{subt}/{f_units}/klines/{symbol.upper()}/{timeframe}/{fname}'
                zf = fetch_file(u, temp_storage, progress_bar=False)
            else:
                logger.info(f'[green]{symbol}[/green] {instr_subtype} {instr_type} parsing data from [red]{fname}[/red]')
            if zf and exists(zf):
                try:
                    data = srows(data, parse_kl_file(zf), keep='last')
                except Exception as err:
                    logger.warning(err)
                    unlink(zf)
        return data, continue_from

    # - load by months
    data, cont_time = _loader(start, stop, 'monthly')

    # - rest data load by days
    if cont_time is not None:
        data_cont, cont_time = _loader(cont_time, stop, 'daily')
        data = srows(data, data_cont, keep='last')

    return data


def update_binance_data_storage(coins=[], quoted_in=['USDT'], market='futures', data_storage=DEFALT_LOCAL_CSV_STORAGE):
    """
    Fetch data from the Binance data storage and save it as local csv files
    TODO: csv is just temporary solution and we need to keep data in DB
    """
    from binance.client import Client, HistoricalKlinesType
    info = get_binance_symbol_info_for_type(['UM', 'CM', 'SPOT'])

    if market.lower() == 'futures':
        for sy in info['binance.um']['symbols']:
            if sy['quoteAsset'] in quoted_in:
                if coins and sy['baseAsset'] not in coins:
                    continue
                symbol = sy['symbol'] 
                start = pd.Timestamp(sy['onboardDate'], unit='ms')
                data = load_binance_kl_history(symbol, start.strftime('%Y-%m-%d'), instr_type='futures', instr_subtype='um')
                # - update in mongo db
                if data is not None and not data.empty:
                    data.to_csv(join(data_storage, 'BINANCE.UM'))
                    # path = f'm1/BINANCEF:{symbol}'
                    # z_del(path, dbname=_DEFAULT_MARKET_DATA_DB)
                    # z_save(path, data, dbname=_DEFAULT_MARKET_DATA_DB)
                    del data

    if market.lower() == 'spot':
        client = Client()
        for sy in info['binance']['symbols']:
            if sy['quoteAsset'] in quoted_in:
                if coins and sy['baseAsset'] not in coins:
                    continue
                symbol = sy['symbol'] 
                # - some dirty way to get historical data start for spot
                d = client.get_historical_klines(
                    symbol, '1M', '2017-01-01', '2100-01-01', klines_type=HistoricalKlinesType.SPOT, limit=1000
                )
                start = pd.Timestamp(d[0][0], unit='ms')
                data = load_binance_kl_history(symbol, start.strftime('%Y-%m-%d'), instr_type='spot', instr_subtype='-')
                # - update in mongo db
                if data is not None and not data.empty:
                    data.to_csv(join(data_storage, 'BINANCE'))
                    # path = f'm1/BINANCE:{symbol}'
                    # z_del(path, dbname=_DEFAULT_MARKET_DATA_DB)
                    # z_save(path, data, dbname=_DEFAULT_MARKET_DATA_DB)
                    del data


def load_binance_markets_info() -> Tuple[pd.DataFrame, Dict[str, pd.DataFrame]]:
    """
    Load binance market info using SPA (non-documented)
    """
    resp = requests.get('https://www.binance.com/bapi/asset/v2/public/asset-service/product/get-products?includeEtf=false')
    data = resp.json()

    market_caps = {}
    m_tags = defaultdict(list)
    for r in data['data']:
        symb = r['s']
        market_caps[symb] = {
            'Symbol': symb, 
            'MarketCap': float(r['cs']) * float(r['c']) / 1_000_000, 
            'Coin': r['b'], 
            'Quoted': r['q'], 
            'Name': r['an'], 
            'Tags': r['tags'], 
        }
        for t in r['tags']:
            m_tags[t].append({
                'Symbol': symb, 
                'MarketCap': float(r['cs']) * float(r['c']) / 1_000_000, 
                'Coin': r['b'], 
                'Quoted': r['q'], 
                'Name': r['an']
        })
    mktcap =  pd.DataFrame.from_dict(market_caps, orient='index').sort_values('MarketCap', ascending=False)

    markets_tags = {}
    for t, m in m_tags.items():
        markets_tags[t.lower() if t else 'none'] = pd.DataFrame.from_records(m).sort_values('MarketCap', ascending=False)

    return mktcap, markets_tags


import re, os
from typing import Callable, List, Union, Optional, Iterable, Any
from os.path import exists, join
import numpy as np
import pandas as pd
import pyarrow as pa
from pyarrow import csv
import psycopg as pg
from functools import wraps

from qubx import logger
from qubx.core.series import TimeSeries, OHLCV, time_as_nsec, Quote, Trade
from qubx.utils.time import infer_series_frequency, handle_start_stop

_DT = lambda x: pd.Timedelta(x).to_numpy().item()
D1, H1 = _DT('1D'), _DT('1h')

DEFAULT_DAILY_SESSION = (_DT('00:00:00.100'), _DT('23:59:59.900'))
STOCK_DAILY_SESSION = (_DT('9:30:00.100'), _DT('15:59:59.900'))
CME_FUTURES_DAILY_SESSION = (_DT('8:30:00.100'), _DT('15:14:59.900'))


def _recognize_t(t: Union[int, str], defaultvalue, timeunit) -> int:
    if isinstance(t, (str, pd.Timestamp)):
        try:
            return np.datetime64(t, timeunit)
        except:
            pass
    return defaultvalue


def _find_column_index_in_list(xs, *args):
    xs = [x.lower() for x in xs]
    for a in args:
        ai = a.lower()
        if ai in xs:
            return xs.index(ai)
    raise IndexError(f"Can't find any from {args} in list: {xs}")


class DataTransformer:

    def __init__(self) -> None:
        self.buffer = []
        self._column_names = []

    def start_transform(self, name: str, column_names: List[str]):
        self._column_names = column_names
        self.buffer = []
    
    def process_data(self, rows_data: Iterable) -> Any:
        if rows_data is not None: 
            self.buffer.extend(rows_data)

    def collect(self) -> Any:
        return self.buffer


class DataReader:

    def get_names(self) -> List[str] :
        raise NotImplemented()

    def read(self, data_id: str, start: str | None=None, stop: str | None=None, 
             transform: DataTransformer = DataTransformer(), 
             chunksize=0, 
             **kwargs
            ) -> Iterable | List:
        raise NotImplemented()


class CsvStorageDataReader(DataReader):
    """
    Data reader for timeseries data stored as csv files in the specified directory
    """

    def __init__(self, path: str) -> None:
        if not exists(path):
            raise ValueError(f"Folder is not found at {path}")
        self.path = path

    def __find_time_idx(self, arr: pa.ChunkedArray, v) -> int:
        ix = arr.index(v).as_py()
        if ix < 0:
            for c in arr.iterchunks():
                a = c.to_numpy()
                ix = np.searchsorted(a, v, side='right')
                if ix > 0 and ix < len(c):
                    ix = arr.index(a[ix]).as_py() - 1
                    break
        return ix

    def __check_file_name(self, name: str) -> str | None:
        _f = join(self.path, name)
        for sfx in ['.csv', '.csv.gz', '']:
            if exists(p:=(_f + sfx)):
                return p 
        return None

    def read(self, data_id: str, start: str | None=None, stop: str | None=None, 
             transform: DataTransformer = DataTransformer(),
             chunksize=0,
             timestamp_formatters = None
            ) -> Iterable | Any:

        f_path = self.__check_file_name(data_id)
        if not f_path:
            ValueError(f"Can't find any csv data for {data_id} in {self.path} !")

        convert_options = None
        if timestamp_formatters is not None:
            convert_options=csv.ConvertOptions(timestamp_parsers=timestamp_formatters)

        table = csv.read_csv(
            f_path, 
            parse_options=csv.ParseOptions(ignore_empty_lines=True),
            convert_options=convert_options
        )
        fieldnames =  table.column_names  

        # - try to find range to load  
        start_idx, stop_idx = 0, table.num_rows
        try:
            _time_field_idx = _find_column_index_in_list(fieldnames, 'time', 'timestamp', 'datetime', 'date')
            _time_type = table.field(_time_field_idx).type
            _time_unit = _time_type.unit if hasattr(_time_type, 'unit') else 's'
            _time_data = table[_time_field_idx]

            # - check if need convert time to primitive types (i.e. Date32 -> timestamp[x])
            _time_cast_function = lambda xs: xs
            if _time_type != pa.timestamp(_time_unit):
                _time_cast_function = lambda xs: xs.cast(pa.timestamp(_time_unit)) 
                _time_data = _time_cast_function(_time_data)

            # - preprocessing start and stop
            t_0, t_1 = handle_start_stop(start, stop, convert=lambda x: _recognize_t(x, None, _time_unit))

            # - check requested range
            if t_0:
                start_idx = self.__find_time_idx(_time_data, t_0)
                if start_idx >= table.num_rows:
                    # no data for requested start date
                    return None

            if t_1:
                stop_idx = self.__find_time_idx(_time_data, t_1)
                if stop_idx < 0 or stop_idx < start_idx:
                    stop_idx = table.num_rows

        except Exception as exc:
            logger.warning(exc)
            logger.info('loading whole file')

        length = (stop_idx - start_idx + 1)
        selected_table = table.slice(start_idx, length)

        # - in this case we want to return iterable chunks of data
        if chunksize > 0:
            def _iter_chunks():
                for n in range(0, length // chunksize + 1):
                    transform.start_transform(data_id, fieldnames)
                    raw_data = selected_table[n*chunksize : min((n+1)*chunksize, length)].to_pandas().to_numpy()
                    transform.process_data(raw_data)
                    yield transform.collect()
            return _iter_chunks()

        transform.start_transform(data_id, fieldnames)
        raw_data = selected_table.to_pandas().to_numpy()
        transform.process_data(raw_data)
        return transform.collect()

    def get_names(self) -> List[str] :
        _n = []
        for s in os.listdir(self.path):
            if (m:=re.match(r'(.*)\.csv(.gz)?$', s)):
                _n.append(m.group(1))
        return _n


class AsPandasFrame(DataTransformer):
    """
    List of records to pandas dataframe transformer
    """

    def start_transform(self, name: str, column_names: List[str]):
        self._time_idx = _find_column_index_in_list(column_names, 'time', 'timestamp', 'datetime', 'date')
        self._column_names = column_names
        self._frame = pd.DataFrame()
    
    def process_data(self, rows_data: Iterable) -> Any:
        self._frame
        p = pd.DataFrame.from_records(rows_data, columns=self._column_names)
        p.set_index(self._column_names[self._time_idx], drop=True, inplace=True)
        p.sort_index(inplace=True)
        self._frame = pd.concat((self._frame, p), axis=0, sort=True)
        return p

    def collect(self) -> Any:
        return self._frame 


class AsOhlcvSeries(DataTransformer):

    def __init__(self, timeframe: str | None = None, timestamp_units='ns') -> None:
        super().__init__()
        self.timeframe = timeframe 
        self._series = None
        self._data_type = None
        self.timestamp_units = timestamp_units

    def start_transform(self, name: str, column_names: List[str]):
        self._time_idx = _find_column_index_in_list(column_names, 'time', 'timestamp', 'datetime', 'date')
        self._volume_idx = None
        self._b_volume_idx = None
        try:
            self._close_idx = _find_column_index_in_list(column_names, 'close')
            self._open_idx = _find_column_index_in_list(column_names, 'open')
            self._high_idx = _find_column_index_in_list(column_names, 'high')
            self._low_idx = _find_column_index_in_list(column_names, 'low')

            try:
                self._volume_idx = _find_column_index_in_list(column_names, 'quote_volume', 'volume', 'vol')
            except: pass

            try:
                self._b_volume_idx = _find_column_index_in_list(column_names, 'taker_buy_volume', 'taker_buy_quote_volume', 'buy_volume')
            except: pass

            self._data_type = 'ohlc'
        except: 
            try:
                self._ask_idx = _find_column_index_in_list(column_names, 'ask')
                self._bid_idx = _find_column_index_in_list(column_names, 'bid')
                self._data_type = 'quotes'
            except: 

                try:
                    self._price_idx = _find_column_index_in_list(column_names, 'price')
                    self._size_idx = _find_column_index_in_list(column_names, 'quote_qty', 'qty', 'size', 'amount', 'volume')
                    self._taker_idx = None
                    try:
                        self._taker_idx = _find_column_index_in_list(column_names, 'is_buyer_maker', 'side', 'aggressive', 'taker', 'is_taker')
                    except: pass

                    self._data_type = 'trades'
                except: 
                    raise ValueError(f"Can't recognize data for update from header: {column_names}")

        self._column_names = column_names
        self._name = name
        if self.timeframe:
            self._series = OHLCV(self._name, self.timeframe)

    def _time(self, t) -> int:
        if self.timestamp_units == 'ns':
            return np.datetime64(t, 'ns').item() 
        return np.datetime64(t, self.timestamp_units).astype('datetime64[ns]').item()

    def _proc_ohlc(self, rows_data: List[List]):
        for d in rows_data:
            self._series.update_by_bar(
                self._time(d[self._time_idx]),
                d[self._open_idx], d[self._high_idx], d[self._low_idx], d[self._close_idx], 
                d[self._volume_idx] if self._volume_idx else 0,
                d[self._b_volume_idx] if self._b_volume_idx else 0
            )

    def _proc_quotes(self, rows_data: List[List]):
        for d in rows_data:
            self._series.update(
                self._time(d[self._time_idx]),
                (d[self._ask_idx] + d[self._bid_idx])/2
            )

    def _proc_trades(self, rows_data: List[List]):
        for d in rows_data:
            a = d[self._taker_idx] if self._taker_idx else 0
            s = d[self._size_idx]
            b = s if a else 0 
            self._series.update(self._time(d[self._time_idx]), d[self._price_idx], s, b)

    def process_data(self, rows_data: List[List]) -> Any:
        if self._series is None:
            ts = [t[self._time_idx] for t in rows_data[:100]]
            self.timeframe = pd.Timedelta(infer_series_frequency(ts)).asm8.item()

            # - create instance after first data received if 
            self._series = OHLCV(self._name, self.timeframe)

        match self._data_type:
            case 'ohlc':
                self._proc_ohlc(rows_data)
            case 'quotes':
                self._proc_quotes(rows_data)
            case 'trades':
                self._proc_trades(rows_data)

        return None

    def collect(self) -> Any:
        return self._series


class AsQuotes(DataTransformer):

    def start_transform(self, name: str, column_names: List[str]):
        self.buffer = list()
        self._time_idx = _find_column_index_in_list(column_names, 'time', 'timestamp', 'datetime')
        self._bid_idx = _find_column_index_in_list(column_names, 'bid')
        self._ask_idx = _find_column_index_in_list(column_names, 'ask')
        self._bidvol_idx = _find_column_index_in_list(column_names, 'bidvol', 'bid_vol', 'bidsize', 'bid_size')
        self._askvol_idx = _find_column_index_in_list(column_names, 'askvol', 'ask_vol', 'asksize', 'ask_size')

    def process_data(self, rows_data: Iterable) -> Any:
        if rows_data is not None: 
            for d in rows_data:
                t = d[self._time_idx] 
                b = d[self._bid_idx]
                a = d[self._ask_idx]
                bv = d[self._bidvol_idx]
                av = d[self._askvol_idx]
                self.buffer.append(Quote(t.as_unit('ns').asm8.item(), b, a, bv, av))


class RestoreTicksFromOHLC(DataTransformer):
    """
    Emulates quotes (and trades) from OHLC bars
    """

    def __init__(self, 
                 trades: bool=False,    # if we also wants 'trades'
                 default_bid_size=1e9,  # default bid/ask is big
                 default_ask_size=1e9,  # default bid/ask is big
                 daily_session_start_end=DEFAULT_DAILY_SESSION,
                 spread=0.0):
        super().__init__()
        self._trades = trades
        self._bid_size = default_bid_size
        self._ask_size = default_ask_size
        self._s2 = spread / 2.0
        self._d_session_start = daily_session_start_end[0]
        self._d_session_end = daily_session_start_end[1]

    def start_transform(self, name: str, column_names: List[str]):
        self.buffer = []
        # - it will fail if receive data doesn't look as ohlcv 
        self._time_idx = _find_column_index_in_list(column_names, 'time', 'timestamp', 'datetime', 'date')
        self._open_idx = _find_column_index_in_list(column_names, 'open')
        self._high_idx = _find_column_index_in_list(column_names, 'high')
        self._low_idx = _find_column_index_in_list(column_names, 'low')
        self._close_idx = _find_column_index_in_list(column_names, 'close')
        self._volume_idx = None
        self._freq = None
        try:
            self._volume_idx = _find_column_index_in_list(column_names, 'volume', 'vol')
        except: pass

        if self._volume_idx is None and self._trades:
            logger.warning("Input OHLC data doesn't contain volume information so trades can't be emulated !")
            self._trades = False

    def process_data(self, rows_data:List[List]) -> Any:
        if rows_data is None: 
            return

        s2 = self._s2

        if self._freq is None:
            ts = [t[self._time_idx] for t in rows_data[:100]]
            self._freq = infer_series_frequency(ts)

            # - timestamps when we emit simulated quotes
            dt = self._freq.astype('timedelta64[ns]').item()
            if dt < D1:
                self._t_start = dt // 10
                self._t_mid1 = dt // 2 - dt // 10
                self._t_mid2 = dt // 2 + dt // 10
                self._t_end = dt - dt // 10
            else:
                self._t_start = self._d_session_start
                self._t_mid1 = dt // 2 - H1
                self._t_mid2 = dt // 2 + H1
                self._t_end = self._d_session_end

        # - input data
        for data in rows_data:
            ti = pd.Timestamp(data[self._time_idx]).as_unit('ns').asm8.item()
            o = data[self._open_idx]
            h=  data[self._high_idx]
            l = data[self._low_idx]
            c = data[self._close_idx]
            rv = data[self._volume_idx] if self._volume_idx else 0

            # - opening quote
            self.buffer.append(Quote(ti + self._t_start, o - s2, o + s2, self._bid_size, self._ask_size))

            if c >= o:
                if self._trades:
                    self.buffer.append(Trade(ti + self._t_start, o - s2, rv * (o - l))) # sell 1
                self.buffer.append(Quote(ti + self._t_mid1, l - s2, l + s2, self._bid_size, self._ask_size))

                if self._trades:
                    self.buffer.append(Trade(ti + self._t_mid1, l + s2, rv * (c - o)))  # buy 1
                self.buffer.append(Quote(ti + self._t_mid2, h - s2, h + s2, self._bid_size, self._ask_size))

                if self._trades:
                    self.buffer.append(Trade(ti + self._t_mid2, h - s2, rv * (h - c)))  # sell 2
            else:
                if self._trades:
                    self.buffer.append(Trade(ti + self._t_start, o + s2, rv * (h - o))) # buy 1
                self.buffer.append(Quote(ti + self._t_mid1, h - s2, h + s2, self._bid_size, self._ask_size))

                if self._trades:
                    self.buffer.append(Trade(ti + self._t_mid1, h - s2, rv * (o - c))) # sell 1
                self.buffer.append(Quote(ti + self._t_mid2, l - s2, l + s2, self._bid_size, self._ask_size))

                if self._trades:
                    self.buffer.append(Trade(ti + self._t_mid2, l + s2, rv * (c - l))) # buy 2

            # - closing quote
            self.buffer.append(Quote(ti + self._t_end, c - s2, c + s2, self._bid_size, self._ask_size))


def _retry(fn):
    @wraps(fn)
    def wrapper(*args, **kw):
        cls = args[0]
        for x in range(cls._reconnect_tries):
            # print(x, cls._reconnect_tries)
            try:
                return fn(*args, **kw)
            except (pg.InterfaceError, pg.OperationalError) as e:
                logger.warning("Database Connection [InterfaceError or OperationalError]")
                # print ("Idle for %s seconds" % (cls._reconnect_idle))
                # time.sleep(cls._reconnect_idle)
                cls._connect()
    return wrapper


class QuestDBConnector(DataReader):
    """
    Very first version of QuestDB connector

    # Connect to an existing QuestDB instance
    >>> db = QuestDBConnector('user=admin password=quest host=localhost port=8812', OhlcvPandasDataProcessor())
    >>> db.read('BINANCEF.ETHUSDT', '2024-01-01')
    """
    _reconnect_tries = 5
    _reconnect_idle = 0.1  # wait seconds before retying

    def __init__(self, connection_url: str) -> None:
        self._connection = None
        self._cursor = None
        self.connection_url = connection_url
        self._connect()

    def _connect(self):
        logger.info("Connecting to QuestDB ...")
        self._connection = pg.connect(self.connection_url, autocommit=True)
        self._cursor = self._connection.cursor()

    @_retry
    def read(self, data_id: str, start: str|None=None, stop: str|None=None, 
             transform: DataTransformer = DataTransformer(),
             chunksize=0,  # TODO: use self._cursor.fetchmany in this case !!!!
             timeframe: str='1m') -> Any:
        start, end = handle_start_stop(start, stop)
        _req = self._prepare_data_sql(data_id, start, end, timeframe)

        self._cursor.execute(_req) # type: ignore
        records = self._cursor.fetchall() # TODO: for chunksize > 0 use fetchmany etc

        names = [d.name for d in self._cursor.description] # type: ignore
        transform.start_transform(data_id, names)

        transform.process_data(records)
        return transform.collect()

    def _prepare_data_sql(self, data_id: str, start: str|None, end: str|None, resample: str) -> str:
        # just a temp hack - actually we need to discuss symbology etc
        symbol = data_id#.split('.')[-1]

        w0 = f"timestamp >= '{start}'" if start else ''
        w1 = f"timestamp <= '{end}'" if end else ''
        where = f'where {w0} and {w1}' if (w0 and w1) else f"where {(w0 or w1)}"

        return f"""
                select timestamp, 
                first(open) as open, 
                max(high) as high,
                min(low) as low,
                last(close) as close,
                sum(volume) as volume,
                sum(quote_volume) as quote_volume,
                sum(count) as count,
                sum(taker_buy_volume) as taker_buy_volume,
                sum(taker_buy_quote_volume) as taker_buy_quote_volume
                from "{symbol.upper()}" {where}
                SAMPLE by {resample};
            """ 

    def _prepare_names_sql(self) -> str:
        return "select table_name from tables()"

    @_retry
    def get_names(self) -> List[str] :
        self._cursor.execute(self._prepare_names_sql()) # type: ignore
        records = self._cursor.fetchall()
        return [r[0] for r in records]

    def __del__(self):
        for c in (self._cursor, self._connection):
            try:
                logger.info("Closing connection")
                c.close()
            except:
                pass


class SnapshotsBuilder(DataTransformer):
    """
    Snapshots assembler from OB updates
    """
    def __init__(self, 
                 levels: int=-1,     # how many levels restore, 1 - TOB, -1 - all
                 as_frame=False      # result is dataframe
    ):
        self.buffer = []
        self.levels = levels
        self.as_frame = as_frame

    def start_transform(self, name: str, column_names: List[str]):
        # initialize buffer / series etc
        # let's keep restored snapshots into some buffer etc
        self.buffer = []

        # do additional init stuff here

    def process_data(self, rows_data:List[List]) -> Any:
        for r in rows_data:
            # restore snapshots and put into buffer or series
            pass

    def collect(self) -> Any:
        # - may be convert it to pandas DataFrame ?
        if self.as_frame:
            return pd.DataFrame.from_records(self.buffer) # or custom transform

        # - or just returns as plain list
        return self.buffer


class QuestDBOrderBookConnector(QuestDBConnector):
    """
    Example of custom OrderBook data connector 
    """

    def _prepare_data_sql(self, data_id: str, start: str|None, end: str|None, resample: str|None) -> str:
        raise NotImplemented("TODO")

    def _prepare_names_sql(self) -> str:
        # return "select table_name from tables() where ..."
        raise NotImplemented("TODO")



from typing import Any, Dict, List, Tuple
from multiprocessing.pool import ThreadPool
import numpy as np
import csv, os

from qubx import logger
from qubx.core.basics import Deal, Position

from qubx.core.series import time_as_nsec
from qubx.core.utils import time_to_str, time_delta_to_str, recognize_timeframe
from qubx.utils.misc import makedirs, Stopwatch

_SW = Stopwatch()


class LogsWriter:
    account_id: str
    strategy_id: str
    run_id: str

    """
    Log writer interface with default implementation
    """
    def __init__(self, account_id: str, strategy_id: str, run_id: str) -> None:
        self.account_id = account_id
        self.strategy_id = strategy_id
        self.run_id = run_id

    def write_data(self, log_type: str, data: List[Dict[str, Any]]):
        pass
    
    def flush_data(self):
        pass


class CsvFileLogsWriter(LogsWriter):
    """
    Simple CSV strategy log data writer. It does data writing in separate thread.
    """
    def __init__(self, account_id: str, strategy_id: str, run_id: str, log_folder='logs') -> None:
        super().__init__(account_id, strategy_id, run_id)

        path = makedirs(log_folder)
        # - it rewrites positions every time
        self._pos_file_path = f"{path}/{self.strategy_id}_{self.account_id}_positions.csv"
        self._balance_file_path = f"{path}/{self.strategy_id}_{self.account_id}_balance.csv"
        _pfl_path = f"{path}/{strategy_id}_{account_id}_portfolio.csv"
        _exe_path =  f"{path}/{strategy_id}_{account_id}_executions.csv"
        self._hdr_pfl = not os.path.exists(_pfl_path)
        self._hdr_exe = not os.path.exists(_exe_path)

        self._pfl_file_ = open(_pfl_path, "+a", newline='')
        self._execs_file_ = open(_exe_path, "+a", newline='')
        self._pfl_writer = csv.writer(self._pfl_file_)
        self._exe_writer = csv.writer(self._execs_file_)
        self.pool = ThreadPool(3)

    @staticmethod
    def _header(d: dict) -> List[str]:
        return list(d.keys()) + ['run_id']

    def _values(self, data: List[Dict[str, Any]]) -> List[List[str]]:
        # - attach run_id (last column)
        return [list((d | {'run_id': self.run_id}).values()) for d in data]

    def _do_write(self, log_type, data):
        match log_type:

            case 'positions':
                with open(self._pos_file_path, "w", newline='') as f:
                    w = csv.writer(f)
                    w.writerow(self._header(data[0]))
                    w.writerows(self._values(data))

            case 'portfolio':
                if self._hdr_pfl:
                    self._pfl_writer.writerow(self._header(data[0]))
                    self._hdr_pfl = False
                self._pfl_writer.writerows(self._values(data))
                self._pfl_file_.flush()

            case 'executions':
                if self._hdr_exe:
                    self._exe_writer.writerow(self._header(data[0]))
                    self._hdr_exe = False
                self._exe_writer.writerows(self._values(data))
                self._execs_file_.flush()

            case 'balance':
                with open(self._balance_file_path, "w", newline='') as f:
                    w = csv.writer(f)
                    w.writerow(self._header(data[0]))
                    w.writerows(self._values(data))

    def write_data(self, log_type: str, data: List[Dict[str, Any]]):
        if len(data) > 0:
            self.pool.apply_async(self._do_write, (log_type, data))

    def flush_data(self):
        try:
            self._pfl_file_.flush()
            self._execs_file_.flush()
        except Exception as e:
            logger.warning(f"Error flushing log writer: {str(e)}")


class _BaseIntervalDumper:
    """
    Basic functionality for all interval based dumpers
    """
    _last_log_time_ns: int
    _freq: np.timedelta64 | None
    
    def __init__(self, frequency: str | None) -> None:
        self._freq: np.timedelta64 | None = recognize_timeframe(frequency) if frequency else None
        self._last_log_time_ns = 0

    def store(self, timestamp: np.datetime64):
        _t_ns = time_as_nsec(timestamp)
        if self._freq:
            _interval_start_time = int(_t_ns  - _t_ns % self._freq)
            if _t_ns - self._last_log_time_ns >= self._freq:
                self.dump(np.datetime64(_interval_start_time, 'ns'), timestamp)
                self._last_log_time_ns = _interval_start_time
        else:
            self.dump(timestamp, timestamp)

    def dump(self, interval_start_time: np.datetime64, actual_timestamp: np.datetime64):
        raise NotImplementedError(f"dump(np.datetime64, np.datetime64) must be implemented in {self.__class__.__name__}")


class PositionsDumper(_BaseIntervalDumper):
    """
    Positions dumper is designed to dump positions once per given interval to storage
    so we could check current situation.
    """
    positions: Dict[str, Position]
    _writer: LogsWriter

    def __init__(
        self, 
        writer: LogsWriter,
        interval: str, 
    ) -> None:
        super().__init__(interval)
        self.positions = dict()
        self._writer = writer

    def attach_positions(self, *positions: Position) -> 'PositionsDumper':
        for p in positions:
            self.positions[p.instrument.symbol] = p
        return self

    def dump(self, interval_start_time: np.datetime64, actual_timestamp: np.datetime64):
        data = []
        for s, p in self.positions.items():
            data.append({
                'timestamp': str(actual_timestamp),
                'instrument_id': s,
                'pnl_quoted': p.total_pnl(),
                'quantity': p.quantity,
                'realized_pnl_quoted': p.r_pnl,
                'avg_position_price': p.position_avg_price if p.quantity != 0.0 else 0.0,
                'current_price': p.last_update_price,
                'market_value_quoted': p.market_value_funds
            })
        self._writer.write_data('positions', data)


class PortfolioLogger(PositionsDumper):
    """
    Portfolio logger - save portfolio records into storage
    """
    def __init__(self, writer: LogsWriter, interval: str) -> None:
        super().__init__(writer, interval)

    def dump(self, interval_start_time: np.datetime64, actual_timestamp: np.datetime64):
        data = []
        for s, p in self.positions.items():
            data.append({
                'timestamp': str(interval_start_time),
                'instrument_id': s,
                'pnl_quoted': p.total_pnl(),
                'quantity': p.quantity,
                'realized_pnl_quoted': p.r_pnl,
                'avg_position_price': p.position_avg_price if p.quantity != 0.0 else 0.0,
                'current_price': p.last_update_price,
                'market_value_quoted': p.market_value_funds,
                'exchange_time': str(actual_timestamp),
                'commissions_quoted': p.commissions,
            })
        self._writer.write_data('portfolio', data)

    def close(self):
        self._writer.flush_data()


class ExecutionsLogger(_BaseIntervalDumper):
    """
    Executions logger - save strategy executions into storage
    """
    _writer: LogsWriter
    _deals: List[Tuple[str, Deal]]

    def __init__(self, writer: LogsWriter, max_records=10) -> None:
        super().__init__(None) # no intervals
        self._writer = writer
        self._max_records = max_records
        self._deals: List[Tuple[str, Deal]] = []
        
    def record_deals(self, symbol: str, deals: List[Deal]):
        for d in deals:
            self._deals.append((symbol, d))
            l_time = d.time

        if len(self._deals) >= self._max_records:
            self.dump(l_time, l_time)

    def dump(self, interval_start_time: np.datetime64, actual_timestamp: np.datetime64):
        data = []
        for s, d in self._deals:
            data.append({
                'timestamp': d.time,
                'instrument_id': s,
                'side': 'buy' if d.amount > 0 else 'sell',
                'filled_qty': d.amount,
                'price': d.price,
                'commissions': d.fee_amount,
                'commissions_quoted': d.fee_currency,
            })
        self._deals.clear()
        self._writer.write_data('executions', data)

    def store(self, timestamp: np.datetime64):
        pass

    def close(self):
        if self._deals:
            t = self._deals[-1][1].time
            self.dump(t, t)
        self._writer.flush_data()


class BalanceLogger(_BaseIntervalDumper):
    """
    Balance logger - send balance on strategy start
    """
    _writer: LogsWriter

    def __init__(self, writer: LogsWriter) -> None:
        super().__init__(None) # no intervals
        self._writer = writer
        
    def record_balance(self, timestamp: np.datetime64, balance: Dict[str, Tuple[float, float]]):
        if balance:
            data = []
            for s, d in balance.items():
                data.append({
                    'timestamp': timestamp,
                    'instrument_id': s,
                    'total': d[0],
                    'locked': d[1],
                })
            self._writer.write_data('balance', data)

    def store(self, timestamp: np.datetime64):
        pass

    def close(self):
        self._writer.flush_data()


class StrategyLogging:
    """
    Just combined loggers functionality
    """
    positions_dumper: PositionsDumper | None = None
    portfolio_logger: PortfolioLogger | None = None
    executions_logger: ExecutionsLogger | None = None
    balance_logger: BalanceLogger | None

    def __init__(
        self, 
        logs_writer: LogsWriter | None = None,
        positions_log_freq: str = '1Min',
        portfolio_log_freq: str = '5Min',
        num_exec_records_to_write = 1      # in live let's write every execution 
    ) -> None:
        # - instantiate loggers
        if logs_writer:
            if positions_log_freq:
                # - store current positions
                self.positions_dumper = PositionsDumper(logs_writer, positions_log_freq)

            if portfolio_log_freq:
                # - store portfolio log
                self.portfolio_logger = PortfolioLogger(logs_writer, portfolio_log_freq)

            # - store executions
            if num_exec_records_to_write >= 1:
                self.executions_logger = ExecutionsLogger(logs_writer, num_exec_records_to_write)

            # - balance logger
            self.balance_logger = BalanceLogger(logs_writer)
        else:
            logger.warning("Log writer is not defined - strategy activity will not be saved !")

    def initialize(self, 
                   timestamp: np.datetime64, 
                   positions: Dict[str, Position],
                   balances: Dict[str, Tuple[float, float]]
    ) -> None:
        # - attach positions to loggers
        if self.positions_dumper:
            self.positions_dumper.attach_positions(*list(positions.values()))

        if self.portfolio_logger:
            self.portfolio_logger.attach_positions(*list(positions.values()))

        # - send balance on start
        if self.balance_logger:
            self.balance_logger.record_balance(timestamp, balances)

    def close(self):
        if self.portfolio_logger:
            self.portfolio_logger.close()

        if self.executions_logger:
            self.executions_logger.close()

    @_SW.watch('loggers')
    def notify(self, timestamp: np.datetime64):
        # - notify position logger
        if self.positions_dumper:
            self.positions_dumper.store(timestamp)

        # - notify portfolio records logger
        if self.portfolio_logger:
            self.portfolio_logger.store(timestamp)

    def save_deals(self, symbol: str, deals: List[Deal]):
        if self.executions_logger:
            self.executions_logger.record_deals(symbol, deals)



import click, sys, yaml, sys, time
from os.path import exists, expanduser
import yaml, configparser, socket

from qubx import lookup, logger, formatter
from qubx.impl.ccxt_connector import CCXTConnector # TODO: need factory !
from qubx.core.strategy import StrategyContext
from qubx.utils.misc import add_project_to_system_path, Struct, logo, version
from qubx.core.loggers import LogsWriter


LOGFILE = 'logs/'


def class_import(name):
    components = name.split('.')
    clz = components[-1]
    mod = __import__('.'.join(components[:-1]), fromlist=[clz])
    mod = getattr(mod, clz)
    return mod


def _instruments_for_exchange(exch: str, symbols: list) -> list:
    instrs = []
    for s in symbols:
        instr = lookup.find_symbol(exch.upper(), s.upper()) 
        if instr is not None:
            instrs.append(instr)
        else:
            logger.warning(f"Can't find instrument for symbol {s} - try to refresh lookup first !")
    return instrs


def load_strategy_config(filename: str) -> Struct:
    with open(filename, 'r') as f:
        content = yaml.safe_load(f)

    config = content['config']
    strat = config['strategy']
    name = strat.split('.')[-1]
    r = Struct(
        strategy = strat,
        name = name,
        parameters = config.get('parameters', dict()),
        connector = config['connector'],
        exchange = config['exchange'],
        account = config.get('account'),
        md_subscr = config['subscription'],
        strategy_trigger = config['trigger'],
        strategy_fit_trigger = config.get('fit', ''),
        portfolio_logger = config.get('logger', None),
        log_positions_interval = config.get('log_positions_interval', None),
        log_portfolio_interval = config.get('log_portfolio_interval', None)
    )

    universe = config['universe']
    if isinstance(universe, dict):
        r.instruments = []
        for e, symbs in universe.items():
            r.instruments.extend(_instruments_for_exchange(e, symbs))
    else:
       r.instruments = _instruments_for_exchange(r.exchange, universe)

    return r


def get_account_config(account_id: str, accounts_cfg_file: str) -> dict | None:
    parser = configparser.ConfigParser()
    try:
        parser.optionxform=str  # type: ignore
        parser.read(accounts_cfg_file)
    except Exception as exc:
        logger.error(f"Can't find { accounts_cfg_file } file for reading {account_id} account info: {str(exc)}")
        return None

    if account_id not in parser:
        logger.error(f"No records for {account_id} found in {accounts_cfg_file} file")
        return None

    cfg = dict(parser[account_id])

    # - check if there any reserved funds
    reserves = {}
    if 'reserves' in cfg: 
        rs = cfg['reserves']
        for r in rs.split(','):
            s,v = r.strip().split(':')
            reserves[s] = float(v)

    # - add account id and reserves
    return cfg | {'account_id': account_id, 'reserves': reserves}


def create_strategy_context(config_file: str, accounts_cfg_file: str, search_paths: list) -> StrategyContext | None:
    cfg = load_strategy_config(config_file)
    try:
        for p in search_paths:
            if exists(pe:=expanduser(p)):
                add_project_to_system_path(pe)
        strategy = class_import(cfg.strategy)
    except Exception as err:
        logger.error(str(err))
        return None

    logger.add(LOGFILE + cfg.name + "_{time}.log", format=formatter, rotation="100 MB", colorize=False)

    # - read account creds
    acc_config = {}
    if cfg.account is not None:
        acc_config = get_account_config(cfg.account, accounts_cfg_file)
        if acc_config is None:
            return None

    # - check connector
    conn = cfg.connector.lower()
    match conn:
        case 'ccxt':
            connector = CCXTConnector(cfg.exchange.lower(), **acc_config)
        case _:
            raise ValueError(f"Connector {conn} is not supported yet !")

    # - generate new run id
    run_id = socket.gethostname() + "-" + str(connector.time().item() // 100_000_000)

    # - get logger
    writer = None
    _w_class = cfg.portfolio_logger
    if _w_class is not None:
        if '.' not in _w_class:
            _w_class = 'qubx.core.loggers.' + _w_class
        try:
            w_class = class_import(_w_class)
            writer = w_class(acc_config['account_id'], strategy.__name__, run_id)
        except Exception as err:
            logger.warning(f"Can't instantiate specified writer {_w_class}: {str(err)}")
            writer = LogsWriter(acc_config['account_id'], strategy.__name__, run_id)
        
    logger.info(f""" - - - <blue>Qubx</blue> (ver. <red>{version()}</red>) - - -\n - Strategy: {strategy}\n - Config: {cfg.parameters} """)
    ctx = StrategyContext(
        strategy, cfg.parameters, connector, connector, 
        instruments        = cfg.instruments, 
        md_subscription    = cfg.md_subscr, 
        trigger_spec       = cfg.strategy_trigger,
        fit_spec           = cfg.strategy_fit_trigger,
        logs_writer        = writer, 
        positions_log_freq = cfg.log_positions_interval,
        portfolio_log_freq = cfg.log_portfolio_interval
    )
 
    return ctx


def _run_in_jupyter(filename: str, accounts: str, paths: list):
    """
    Helper for run this in jupyter console
    """
    try:
        from jupyter_console.app import ZMQTerminalIPythonApp
    except ImportError:
        logger.error("Can't find <red>ZMQTerminalIPythonApp</red> module - try to install jupyter package first")
        return 
    try:
        import nest_asyncio
    except ImportError:
        logger.error("Can't find <red>nest_asyncio</red> module - try to install it first")
        return 

    class TerminalRunner(ZMQTerminalIPythonApp):
        def __init__(self, **kwargs) -> None:
            self.init_code = kwargs.pop('init_code')
            super().__init__(**kwargs)
        def init_banner(self):
            pass
        def initialize(self, argv=None):
            super().initialize(argv=[])
            self.shell.run_cell(self.init_code)

    logger.info("Running in Jupyter console")
    TerminalRunner.launch_instance(init_code=f"""
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
import qubx
%qubxd
import pandas as pd
import nest_asyncio; nest_asyncio.apply()
from qubx.utils.misc import dequotify, quotify
from qubx.utils.runner import create_strategy_context
from qubx.pandaz.utils import *
import qubx.pandaz.ta as pta

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
ctx = create_strategy_context('{filename}', '{accounts}', {paths})
ctx.start()

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
def orders(symbol=None):
    return ctx.exchange_service.get_orders(symbol)

def trade(symbol, qty, price=None, tif='gtc'):
    return ctx.trade(symbol, qty, price, tif)

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
def pnl_report(all=True):
    from tabulate import tabulate
    d = dict()
    for s, p in ctx.positions.items():
        mv = round(p.market_value_funds, 3)
        if mv != 0.0 or all:
            d[dequotify(s)] = dict(
                Position=round(p.quantity, p.instrument.size_precision),  
                PnL=p.total_pnl(), 
                AvgPrice=round(p.position_avg_price_funds, p.instrument.price_precision), 
                LastPrice=round(p.last_update_price, p.instrument.price_precision),
                MktValue=mv
            )
    d = pd.DataFrame.from_dict(d).T
    # d = d[d['PnL'] != 0.0]
    if d.empty:
        print('-(no open positions yet)-')
        return
    d = d.sort_values('PnL' ,ascending=False)
    # d = pd.concat((d, pd.Series(dict(TOTAL=d['PnL'].sum()), name='PnL'))).fillna('')
    d = pd.concat((d, scols(pd.Series(dict(TOTAL=d['PnL'].sum()), name='PnL'), pd.Series(dict(TOTAL=d['MktValue'].sum()), name='MktValue')))).fillna('')
    print(tabulate(d, ['Position', 'PnL', 'AvgPrice', 'LastPrice', 'MktValue'], tablefmt='rounded_grid'))
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
__exit = exit
def exit():
    ctx.stop(); __exit()
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
""")


@click.command()
@click.argument('filename', type=click.Path(exists=True))
@click.option('--accounts', '-a', default='accounts.cfg', type=click.STRING, help='Live accounts configuration file')
@click.option('--paths', '-p', multiple=True, default=['../', '~/projects/'], type=click.STRING, help='Live accounts configuration file')
@click.option('--jupyter', '-j', is_flag=True, default=False, help='Run strategy in jupyter console', show_default=True)
def run(filename: str, accounts: str, paths: list, jupyter: bool):
    if jupyter:
        _run_in_jupyter(filename, accounts, paths)
        return

    # - show Qubx logo with current version
    logo()

    # - create context
    ctx = create_strategy_context(filename, accounts, paths)
    if ctx is None:
        return

    # - run main loop
    try:
        ctx.start()

        # - just wake up every 60 sec and check if it's OK 
        while True: 
            time.sleep(60)

    except KeyboardInterrupt:
        ctx.stop()
        time.sleep(1)
        sys.exit(0)


if __name__ == "__main__":
    run()




from typing import Iterable, List, Set, Tuple, Union, Dict, Optional
from dataclasses import dataclass
import numpy as np
import pandas as pd

from qubx import logger
from qubx.core.basics import Position
from qubx.core.strategy import StrategyContext, PositionsTracker


@dataclass
class Capital:
    capital: float
    released_amount: float
    symbols_to_close: List[str] | None = None


class PortfolioRebalancerTracker(PositionsTracker):
    """
    Simple portfolio rebalancer based on fixed weights
    """
    capital_invested: float
    tolerance: float

    def __init__(self, ctx: StrategyContext, capital_invested: float, tolerance: float) -> None:
        super().__init__(ctx)
        self.capital_invested = capital_invested
        self.tolerance = tolerance
        if self.capital_invested == 0: 
            self.capital_invested = ctx.get_capital()

    def calculate_released_capital(self, symbols_to_close: List[str] | None = None) -> Tuple[float, List[str]]:
        """
        Calculate capital that would be released if close positions for provided symbols_to_close list
        """
        released_capital_after_close = 0.0
        closed_symbols = []
        if  symbols_to_close is not None:
            for symbol in symbols_to_close:
                p = self.ctx.positions.get(symbol)
                if p is not None and p.quantity != 0:
                    released_capital_after_close += p.get_amount_released_funds_after_closing(to_remain=self.ctx.get_reserved(p.instrument))
                    closed_symbols.append(symbol)
        return released_capital_after_close, closed_symbols

    def estimate_capital_to_trade(self, symbols_to_close: List[str] | None = None) -> Capital:
        released_capital = 0.0
        closed_positions = None

        if symbols_to_close is not None:
            released_capital, closed_positions = self.calculate_released_capital(symbols_to_close)

        cap_to_invest = self.ctx.get_capital() + released_capital
        if self.capital_invested > 0:
            cap_to_invest = min(self.capital_invested, cap_to_invest) 

        return Capital(cap_to_invest, released_capital, closed_positions)

    def process_signals(self, signals: pd.DataFrame):
        """
        TODO: will be moved to StrategyContext in future - temporary here

        signals: dataframe with desired position sizes
        """
        last_signal = dict(signals.iloc[-1])
        to_close = [symbol  for symbol, new_size in last_signal.items() if new_size == 0]
        to_open = {symbol: new_size  for symbol, new_size in last_signal.items() if new_size != 0}

        # close positions first - we need to release capital 
        for s in to_close:
            if (pos := self.ctx.positions.get(s)):
                reserved = self.ctx.get_reserved(pos.instrument)
                to_close = self._how_much_can_be_closed(pos.quantity, reserved)
                logger.info(f"(PortfolioRebalancerTracker) {s} - closing {to_close} from {pos.quantity} amount (reserved: {reserved})")
                try:
                    self.ctx.trade(s, -to_close)
                except Exception as err:
                    logger.error(f"(PortfolioRebalancerTracker) {s} Error processing closing order: {str(err)}")
            else:
                logger.error(f"(PortfolioRebalancerTracker) Position for {s} is required to be closed but can't be found in context !")

        # open or alter or open new positions
        for s, n in to_open.items():
            if (pos := self.ctx.positions.get(s)):
                trade_size = (n - pos.quantity)
                trade_size_change_pct = abs(trade_size / pos.quantity) if pos.quantity != 0 else 1
                if 100 * trade_size_change_pct > self.tolerance: 
                    logger.info(f"(PortfolioRebalancerTracker) {s} - change position {pos.quantity} -> {n} (tolerance: {self.tolerance}%)")
                    try:
                        self.ctx.trade(s, trade_size)
                    except Exception as err:
                        logger.error(f"(PortfolioRebalancerTracker) {s} Error processing opening order: {str(err)}")
                else:
                    logger.info(f"(PortfolioRebalancerTracker) {s} - position change ({pos.quantity} -> {n}) is smaller than tolerance {self.tolerance}%")
                    
            else:
                logger.error(f"(PortfolioRebalancerTracker) Position for {s} is required to be changed but can't be found in context !")

    def close_all(self):
        for s, pos in self.ctx.positions.items():
            if pos.quantity != 0:
                reserved = self.ctx.get_reserved(pos.instrument)
                to_close = self._how_much_can_be_closed(pos.quantity, reserved)
                if to_close != 0:
                    try:
                        logger.info(f"(PortfolioRebalancerTracker) {s} - closing {to_close} from {pos.quantity} amount (reserved: {reserved})")
                        self.ctx.trade(s, -pos.quantity)
                    except Exception as err:
                        logger.error(f"(PortfolioRebalancerTracker) {s} Error processing closing order: {str(err)}")

    def _how_much_can_be_closed(self, position: float, to_remain: float) -> float:
        d = np.sign(position)
        qty_to_close = position
        if to_remain != 0 and position != 0 and np.sign(to_remain) == d:
            qty_to_close = max(position - to_remain, 0) if d > 0 else min(position - to_remain, 0)
        return qty_to_close



import numpy as np
cimport numpy as np
from collections import deque

from qubx.core.series cimport TimeSeries, Indicator, RollingSum, nans


cdef extern from "math.h":
    float INFINITY


cdef class Sma(Indicator):
    """
    Simple Moving Average indicator
    """
    cdef unsigned int period
    cdef RollingSum summator

    """
    Simple moving average
    """
    def __init__(self, str name, TimeSeries series, int period):
        self.period = period
        self.summator = RollingSum(period)
        super().__init__(name, series)

    cpdef double calculate(self, long long time, double value, short new_item_started):
        cdef double r = self.summator.update(value, new_item_started)
        return np.nan if self.summator.is_init_stage else r / self.period


def sma(series:TimeSeries, period: int): 
    return Sma.wrap(series, period)


cdef class Ema(Indicator):
    """
    Exponential moving average
    """
    cdef int period
    cdef np.ndarray __s
    cdef int __i
    cdef double alpha
    cdef double alpha_1
    cdef unsigned short init_mean 
    cdef unsigned short _init_stage

    def __init__(self, str name, TimeSeries series, int period, init_mean=True):
        self.period = period

        # when it's required to initialize this ema by mean on first period
        self.init_mean = init_mean
        if init_mean:
            self.__s = nans(period)
            self.__i = 0

        self._init_stage = 1
        self.alpha = 2.0 / (1.0 + period)
        self.alpha_1 = (1 - self.alpha)
        super().__init__(name, series)

    cpdef double calculate(self, long long time, double value, short new_item_started):
        cdef int prev_bar_idx = 0 if new_item_started else 1 

        if self._init_stage:
            if np.isnan(value): return np.nan

            if new_item_started:
                self.__i += 1
                if self.__i > self.period - 1:
                    self._init_stage = False
                    return self.alpha * value + self.alpha_1 * self[prev_bar_idx]

            if self.__i == self.period - 1:
                self.__s[self.__i] = value 
                return np.nansum(self.__s) / self.period

            self.__s[self.__i] = value 
            return np.nan

        if len(self) == 0:
            return value

        return self.alpha * value + self.alpha_1 * self[prev_bar_idx]


def ema(series:TimeSeries, period: int, init_mean: bool = True):
    return Ema.wrap(series, period, init_mean=init_mean)


cdef class Tema(Indicator):
    cdef int period
    cdef unsigned short init_mean 
    cdef TimeSeries ser0
    cdef Ema ema1
    cdef Ema ema2
    cdef Ema ema3

    def __init__(self, str name, TimeSeries series, int period, init_mean=True):
        self.period = period
        self.init_mean = init_mean
        self.ser0 = TimeSeries('ser0', series.timeframe, series.max_series_length)
        self.ema1 = ema(self.ser0, period, init_mean)
        self.ema2 = ema(self.ema1, period, init_mean)
        self.ema3 = ema(self.ema2, period, init_mean)
        super().__init__(name, series)
        
    cpdef double calculate(self, long long time, double value, short new_item_started):
        self.ser0.update(time, value)
        return 3 * self.ema1[0] - 3 * self.ema2[0] + self.ema3[0]


def tema(series:TimeSeries, period: int, init_mean: bool = True):
    return Tema.wrap(series, period, init_mean=init_mean)


cdef class Dema(Indicator):
    cdef int period
    cdef unsigned short init_mean 
    cdef TimeSeries ser0
    cdef Ema ema1
    cdef Ema ema2

    def __init__(self, str name, TimeSeries series, int period, init_mean=True):
        self.period = period
        self.init_mean = init_mean
        self.ser0 = TimeSeries('ser0', series.timeframe, series.max_series_length)
        self.ema1 = ema(self.ser0, period, init_mean)
        self.ema2 = ema(self.ema1, period, init_mean)
        super().__init__(name, series)
        
    cpdef double calculate(self, long long time, double value, short new_item_started):
        self.ser0.update(time, value)
        return 2 * self.ema1[0] - self.ema2[0]


def dema(series:TimeSeries, period: int, init_mean: bool = True):
    return Dema.wrap(series, period, init_mean=init_mean)


cdef class Kama(Indicator):
    cdef int period
    cdef int fast_span
    cdef int slow_span
    cdef double _S1 
    cdef double _K1 
    cdef _x_past
    cdef RollingSum summator

    def __init__(self, str name, TimeSeries series, int period, int fast_span=2, int slow_span=30):
        self.period = period
        self.fast_span = fast_span
        self.slow_span = slow_span
        self._S1 = 2.0 / (slow_span + 1)
        self._K1 = 2.0 / (fast_span + 1) - self._S1
        self._x_past = deque(nans(period+1), period+1)
        self.summator = RollingSum(period)
        super().__init__(name, series)

    cpdef double calculate(self, long long time, double value, short new_item_started):
        if new_item_started:
            self._x_past.append(value)
        else:
            self._x_past[-1] = value

        cdef double rs = self.summator.update(abs(value - self._x_past[-2]), new_item_started)
        cdef double er = abs(value - self._x_past[0]) / rs
        cdef double sc = (er * self._K1 + self._S1) ** 2

        if self.summator.is_init_stage:
            if not np.isnan(self._x_past[1]):
                return value
            return np.nan

        return sc * value + (1 - sc) * self[0 if new_item_started else 1]


def kama(series:TimeSeries, period: int, fast_span:int=2, slow_span:int=30):
    return Kama.wrap(series, period, fast_span, slow_span)


cdef class Highest(Indicator):
    cdef int period
    cdef queue

    def __init__(self, str name, TimeSeries series, int period):
        self.period = period
        self.queue = deque([np.nan] * period, maxlen=period)
        super().__init__(name, series)

    cpdef double calculate(self, long long time, double value, short new_item_started):
        """
        Not a most effictive algo but simplest and can handle updated last value
        """
        cdef float r = np.nan

        if not np.isnan(value):
            if new_item_started:
                self.queue.append(value)
            else:
                self.queue[-1] = value

        if not np.isnan(self.queue[0]):
            r = max(self.queue) 

        return r


def highest(series:TimeSeries, period:int):
    return Highest.wrap(series, period)


cdef class Lowest(Indicator):
    cdef int period
    cdef queue

    def __init__(self, str name, TimeSeries series, int period):
        self.period = period
        self.queue = deque([np.nan] * period, maxlen=period)
        super().__init__(name, series)

    cpdef double calculate(self, long long time, double value, short new_item_started):
        """
        Not a most effictive algo but simplest and can handle updated last value
        """
        cdef float r = np.nan

        if not np.isnan(value):
            if new_item_started:
                self.queue.append(value)
            else:
                self.queue[-1] = value

        if not np.isnan(self.queue[0]):
            r = min(self.queue) 

        return r


def lowest(series:TimeSeries, period:int):
    return Lowest.wrap(series, period)


# - - - - TODO !!!!!!!
cdef class Std(Indicator):
    cdef int period

    def __init__(self, str name, TimeSeries series, int period):
        self.period = period
        super().__init__(name, series)

    cpdef double calculate(self, long long time, double value, short new_item_started):
        pass


def std(series:TimeSeries, period:int, mean=0):
    return Std.wrap(series, period)
# - - - - TODO !!!!!!!



from typing import Any, Optional, List, Dict, Tuple
from collections import defaultdict

import numpy as np

from qubx import lookup, logger
from qubx.core.basics import Instrument, Position, TransactionCostsCalculator, dt_64, Deal, Order


class AccountProcessor:
    """
    Account processor class
    """
    account_id: str
    base_currency: str
    reserved: Dict[str, float]                         # how much asset is reserved against the trading 
    _balances: Dict[str, Tuple[float, float]]
    _active_orders: Dict[str|int, Order]               # active orders
    _processed_trades: Dict[str|int, List[str|int]] 
    _positions: Dict[str, Position]
    _total_capital_in_base: float = 0.0
    _locked_capital_in_base: float = 0.0
    _locked_capital_by_order: Dict[str|int, float]

    def __init__(self, 
                 account_id: str, 
                 base_currency: str, 
                 reserves: Dict[str, float] | None, 
                 total_capital: float=0, 
                 locked_capital: float=0
    ) -> None:
        self.account_id = account_id
        self.base_currency = base_currency
        self.reserved = dict() if reserves is None else reserves
        self._processed_trades = defaultdict(list)
        self._active_orders = dict()
        self._positions = {}
        self._locked_capital_by_order = dict()
        self._balances = dict()
        self.update_base_balance(total_capital, locked_capital)

    def update_base_balance(self, total_capital: float, locked_capital: float):
        """
        Update base currency balance
        """
        self._total_capital_in_base = total_capital
        self._locked_capital_in_base = locked_capital

    def update_balance(self, symbol: str, total_capital: float, locked_capital: float):
        self._balances[symbol] = (total_capital, locked_capital)

    def get_balances(self) -> Dict[str, Tuple[float, float]]:
        return dict(self._balances)

    def attach_positions(self, *position: Position) -> 'AccountProcessor':
        for p in position:
            self._positions[p.instrument.symbol] = p 
        return self

    def update_position_price(self, time: dt_64, symbol: str, price: float):
        p = self._positions[symbol]
        p.update_market_price(time, price, 1)

    def get_capital(self) -> float:
        # TODO: need to take in account leverage and funds currently locked 
        return self._total_capital_in_base - self._locked_capital_in_base  

    def get_reserved(self, instrument: Instrument) -> float:
        """
        Check how much were reserved for this instrument
        """
        return self.reserved.get(instrument.symbol, self.reserved.get(instrument.base, 0))

    def process_deals(self, symbol: str, deals: List[Deal]):
        pos = self._positions.get(symbol)

        if pos is not None:
            conversion_rate = 1
            instr = pos.instrument
            traded_amnt, realized_pnl, deal_cost = 0, 0, 0

            # - check if we need conversion rate for this instrument
            # - TODO - need test on it !
            if instr._aux_instrument is not None:
                aux_pos = self._positions.get(instr._aux_instrument.symbol)
                if aux_pos:
                    conversion_rate = aux_pos.last_update_price
                else:
                    logger.error(f"Can't find additional instrument {instr._aux_instrument} for estimating {symbol} position value !!!") 

            # - process deals
            for d in deals:
                if d.id not in self._processed_trades[d.order_id]: 
                    self._processed_trades[d.order_id].append(d.id)
                    realized_pnl += pos.update_position_by_deal(d, conversion_rate) 
                    deal_cost += d.amount * d.price / conversion_rate
                    traded_amnt += d.amount
                    logger.info(f"  ::  traded {d.amount} for {symbol} @ {d.price} -> {realized_pnl:.2f}")
                    self._total_capital_in_base -= deal_cost

    def _lock_limit_order_value(self, order: Order) -> float:
        pos = self._positions.get(order.symbol)
        excess = 0.0
        # - we handle only instruments it;s subscribed to
        if pos:
            sgn = -1 if order.side == 'SELL' else +1
            pos_change = sgn * order.quantity
            direction = np.sign(pos_change)
            prev_direction = np.sign(pos.quantity)
            # how many shares are closed/open
            qty_closing = min(abs(pos.quantity), abs(pos_change)) * direction if prev_direction != direction else 0
            qty_opening = pos_change if prev_direction == direction else pos_change - qty_closing
            excess = abs(qty_opening) * order.price

            if excess > 0:
                self._locked_capital_in_base += excess
                self._locked_capital_by_order[order.id] = excess

        return excess

    def _unlock_limit_order_value(self, order: Order):
        if order.id in self._locked_capital_by_order:
            excess = self._locked_capital_by_order.pop(order.id )
            self._locked_capital_in_base -= excess

    def process_order(self, order: Order):
        _new = order.status == 'NEW'
        _open = order.status == 'OPEN'
        _closed = order.status == 'CLOSED'
        _cancel = order.status == 'CANCELED'

        if _open or _new:
            self._active_orders[order.id] = order

            # - calculate amount locked by this order
            if order.type == 'LIMIT':
                self._lock_limit_order_value(order)

        if _closed or _cancel:
            if order.id in self._processed_trades:
                self._processed_trades.pop(order.id)

            if order.id in self._active_orders:
                self._active_orders.pop(order.id)

        # - calculate amount to unlock after canceling
        if _cancel and order.type == 'LIMIT':
            self._unlock_limit_order_value(order)

        logger.info(f"Order {order.id} {order.type} {order.side} {order.quantity} of {order.symbol} -> {order.status}")

    def add_active_orders(self, orders: Dict[str, Order]):
        for oid, od in orders.items():
            self._active_orders[oid] = od

    def get_orders(self, symbol: str | None = None) -> List[Order]:
        ols = list(self._active_orders.values())
        if symbol is not None:
            ols = list(filter(lambda x: x.symbol == symbol, ols))
        return ols

    def positions_report(self) -> dict:
        rep = {}
        for p in self._positions.values():
            rep[p.instrument.symbol] = {'Qty': p.quantity, 'Price': p.position_avg_price_funds, 'PnL': p.pnl, 'MktValue': p.market_value_funds}
        return rep


"""
 # All interfaces related to strategy etc
"""
from typing import Any, Callable, Dict, List, Optional, Tuple, Union
from types import FunctionType
from collections import defaultdict
from dataclasses import dataclass
from threading import Thread
from multiprocessing.pool import ThreadPool
import traceback

import pandas as pd

from qubx import lookup, logger
from qubx.core.account import AccountProcessor
from qubx.core.helpers import BasicScheduler, CachedMarketDataHolder, process_schedule_spec
from qubx.core.loggers import LogsWriter, StrategyLogging
from qubx.core.lookups import InstrumentsLookup
from qubx.core.basics import Deal, Instrument, Order, Position, Signal, dt_64, td_64, CtrlChannel
from qubx.core.series import TimeSeries, Trade, Quote, Bar, OHLCV
from qubx.utils.misc import Stopwatch
from qubx.utils.time import convert_seconds_to_str


@dataclass
class TriggerEvent:
    time: dt_64
    type: str
    instrument: Optional[Instrument]
    data: Optional[Any] 


class IDataProvider:

    def subscribe(self, subscription_type: str, symbols: List[str], **kwargs) -> bool:
        raise NotImplementedError("subscribe")

    def get_communication_channel(self) -> CtrlChannel:
        raise NotImplementedError("get_communication_channel")

    def get_historical_ohlcs(self, symbol: str, timeframe: str, nbarsback: int) -> List[Bar]:
        raise NotImplementedError("get_historical_ohlcs")

    def get_quote(self, symbol: str) -> Quote | None:
        raise NotImplementedError("get_quote")


class IExchangeServiceProvider:
    acc: AccountProcessor

    def time(self) -> dt_64:
        """
        Returns current time
        """
        raise NotImplementedError("time is not implemented")

    def get_account(self) -> AccountProcessor:
        return self.acc

    def get_name(self) -> str:
        raise NotImplementedError("get_name is not implemented")

    def get_capital(self) -> float:
        return self.acc.get_capital()

    def send_order(self, instrument: Instrument, order_side: str, order_type: str, amount: float, price: float | None = None, 
        client_id: str | None = None, time_in_force: str='gtc') -> Order | None:
        raise NotImplementedError("send_order is not implemented")

    def cancel_order(self, order_id: str) -> Order | None:
        raise NotImplementedError("cancel_order is not implemented")

    def get_orders(self, symbol: str | None = None) -> List[Order]:
        raise NotImplementedError("get_orders is not implemented")

    def get_position(self, instrument: Instrument) -> Position:
        raise NotImplementedError("get_position is not implemented")

    def get_base_currency(self) -> str:
        raise NotImplementedError("get_basic_currency is not implemented")


class PositionsTracker:
    ctx: 'StrategyContext'

    def __init__(self, ctx: 'StrategyContext') -> None:
        self.ctx = ctx
    

class IStrategy:
    ctx: 'StrategyContext'

    def on_start(self, ctx: 'StrategyContext'):
        """
        This method is called strategy is started
        """
        pass

    def on_fit(self, ctx: 'StrategyContext', fit_time: str | pd.Timestamp, previous_fit_time: str | pd.Timestamp | None = None):
        """
        This method is called when it's time to fit model
        :param fit_time: last time of fit data to use 
        :param previous_fit_time: last time of fit data used in previous fit call 
        """
        return None

    def on_event(self, ctx: 'StrategyContext', event: TriggerEvent) -> Optional[List[Signal]]:
        return None

    def on_stop(self, ctx: 'StrategyContext'):
        pass

    def tracker(self, ctx: 'StrategyContext') -> PositionsTracker | None:
        pass


def _dict_with_exception(dct, f):
    if f not in dct:
        raise ValueError(f"Configuration {dct} must contain field '{f}'")
    return dct[f]


def _round_down_at_min_qty(x: float, min_size: float) -> float:
    return (int(x / min_size)) * min_size


_SW = Stopwatch()

 
class StrategyContext:
    MAX_NUMBER_OF_STRATEGY_FAILURES = 10

    strategy: IStrategy                         # strategy instance
    exchange_service: IExchangeServiceProvider  # service for exchange API: orders managemewnt
    data_provider: IDataProvider                # market data provider 
    instruments: List[Instrument]               # list of instruments this strategy trades
    positions: Dict[str, Position]              # positions of the strategy (instrument -> position)

    # - loggers
    _logging: StrategyLogging                    # recording all activities for the strat: execs, positions, portfolio

    # - cached marked data anb scheduler
    _cache: CachedMarketDataHolder # market data cache
    _scheduler: BasicScheduler

    # - configuration
    _market_data_subcription_type:str = 'unknown'
    _market_data_subcription_params: dict = dict()
    _thread_data_loop: Thread | None = None            # market data loop

    _trig_interval_in_bar_nsec: int
    _trig_bar_freq_nsec: int
    _trig_on_bar: bool = False
    _trig_on_time: bool = False
    _trig_on_quote: bool = False
    _trig_on_trade: bool = False
    _trig_on_book: bool = False
    _current_bar_trigger_processed: bool = False
    _is_initilized: bool = False
    _symb_to_instr: Dict[str, Instrument]
    __strategy_id: str
    __order_id: int 
    __handlers: Dict[str, Callable[['StrategyContext', str, Any], TriggerEvent | None]]
    __fit_is_running: bool = False                       # during fitting working it stops calling on_event method
    __init_fit_was_called: bool = False                  # true if initial fit was already run
    __init_fit_args: Tuple = (None, None)                # arguments for initial on_fit() method call
    __pool: ThreadPool | None = None                     # thread pool used for running aux tasks (fit etc)

    def __init__(self, 
            # - strategy with parameters
            strategy: IStrategy, config: Dict[str, Any] | None,
            # - - - - - - - - - - - - - - - - - - - - -

            # - data provider and exchange service
            data_provider: IDataProvider,
            exchange_service: IExchangeServiceProvider, 
            instruments: List[Instrument],
            # - - - - - - - - - - - - - - - - - - - - -

            # - data subscription - - - - - - - - - - -
            md_subscription: Dict[str,Any] = dict(type='ohlc', timeframe='1Min', nback=60),
            # - - - - - - - - - - - - - - - - - - - - -

            # - when need to trigger and fit strategy - - - - - - -
            trigger_spec: str = 'bar: -1Sec',       # 1 sec before subscription bar is closed
            fit_spec: str | None = None,
            # - - - - - - - - - - - - - - - - - - - - -

            # - how to write logs - - - - - - - - - -
            logs_writer: LogsWriter | None = None,
            positions_log_freq: str = '1Min',
            portfolio_log_freq: str = '5Min',
            num_exec_records_to_write = 1      # in live let's write every execution 
        ) -> None:

        # - initialization
        self.exchange_service = exchange_service
        self.data_provider = data_provider
        self.config = config
        self.instruments = instruments
        self.positions = {}
        self.__fit_is_running = False         
        self.__init_fit_was_called = False     
        self.__pool = None 

        # - for fast access to instrument by it's symbol
        self._symb_to_instr = {i.symbol: i for i in instruments}

        # - create scheduler
        self._scheduler = BasicScheduler(self.data_provider.get_communication_channel(), lambda: self.time().item())

        # - instantiate logging functional
        self._logging = StrategyLogging(logs_writer, positions_log_freq, portfolio_log_freq, num_exec_records_to_write)

        # - extract data and event handlers
        self.__handlers = {
            n.split('_processing_')[1]: f for n, f in self.__class__.__dict__.items() 
            if type(f) == FunctionType and n.startswith('_processing_') 
        }

        # - create strategy instance and populate custom paramameters
        self._instantiate_strategy(strategy, config)

        # - process market data configuration
        self.__check_how_to_listen_to_market_data(md_subscription)
 
        # - process trigger and fit configurations
        self.__check_how_to_trigger_and_fit_strategy(trigger_spec, fit_spec)

        # - run cron scheduler
        self._scheduler.run()

    def _instantiate_strategy(self, strategy: IStrategy, config: Dict[str, Any] | None):
        # - set parameters to strategy (not sure we will do it here)
        self.strategy = strategy
        if isinstance(strategy, type):
            self.strategy = strategy()
        self.strategy.ctx = self

        # TODO: - trackers - - - - - - - - - - - - -
        # - here we need to instantiate trackers 
        # - need to think how to do it properly !!!
        # TODO: - trackers - - - - - - - - - - - - -

        # - set strategy custom parameters 
        self.populate_parameters_to_strategy(self.strategy, **config if config else {})
        self._is_initilized = False
        self.__strategy_id = self.strategy.__class__.__name__ + "_"
        self.__order_id = self.time().item() // 100_000_000

    def __check_how_to_listen_to_market_data(self, md_config: dict):
        self._market_data_subcription_type = _dict_with_exception(md_config, 'type').lower()
        match self._market_data_subcription_type:
            case 'ohlc':
                timeframe = _dict_with_exception(md_config, 'timeframe')
                self._market_data_subcription_params = {
                    'timeframe': timeframe,
                    'nback': md_config.get('nback', 1), 
                }
                self._cache = CachedMarketDataHolder(timeframe) 

            case 'trade' | 'trades' | 'tas':
                self._cache = CachedMarketDataHolder('1Sec') 

            case 'quote' | 'quotes':
                self._cache = CachedMarketDataHolder('1Sec') 

            case 'ob' | 'orderbook':
                self._cache = CachedMarketDataHolder('1Sec') 

            case _:
                raise ValueError(f"{self._market_data_subcription_type} is not a valid value for market data subcription type !!!")

    def __check_how_to_trigger_and_fit_strategy(self, trigger_schedule: str | None, fit_schedue: str | None):
        _td2ns = lambda x: x.as_unit('ns').asm8.item()

        self._trig_interval_in_bar_nsec = 0
        
        if not trigger_schedule:
            raise ValueError("trigger parameter can't be empty !")
        t_rules = process_schedule_spec(trigger_schedule)
        f_rules = process_schedule_spec(fit_schedue)

        if t_rules is None:
            raise ValueError(f"Couldn't recognize 'trigger' parameter specification: {trigger_schedule} !")

        # - we can use it as reference if bar timeframe is not secified
        _default_data_timeframe = pd.Timedelta(self._cache.default_timeframe)

        # - check trigger spec - - - - -
        match t_rules['type']:
            # it triggers on_event method when new bar is formed.
            # in this case it won't arm scheduler but just ruled by actual market data updates 
            # - TODO: so probably need to drop this in favor to cron tasks
            # it's also possible to specify delay
            # "bar: -1s"      - it uses default timeframe and wake up 1 sec before every bar's close
            # "bar.5m: -5sec" - 5 sec before 5min bar's close
            # "bar.5m: 1sec"  - 1 sec after 5min bar closed (in next bar)
            case 'bar':
                _r_tf = t_rules.get('timeframe')
                _bar_timeframe = _default_data_timeframe if not _r_tf else pd.Timedelta(_r_tf)
                _inside_bar_delay: pd.Timedelta = t_rules.get('delay', pd.Timedelta(0))

                if abs(pd.Timedelta(_inside_bar_delay)) > pd.Timedelta(_bar_timeframe):
                    raise ValueError(f"Delay must be less or equal to bar's timeframe for bar trigger: you used {_inside_bar_delay} delay for {_bar_timeframe}")

                # for positive delay - trigger strategy when this interval passed after new bar's open
                if _inside_bar_delay >= pd.Timedelta(0): 
                    self._trig_interval_in_bar_nsec = _td2ns(_inside_bar_delay)

                # for negative delay - trigger strategy when time is closer to bar's closing time more than this interval
                else:
                    self._trig_interval_in_bar_nsec = _td2ns(_bar_timeframe + _inside_bar_delay)

                self._trig_bar_freq_nsec = _td2ns(_bar_timeframe)
                self._trig_on_bar = True

                logger.debug(f"Triggering strategy on every {convert_seconds_to_str(self._trig_bar_freq_nsec/1e9)} bar after {convert_seconds_to_str(self._trig_interval_in_bar_nsec/1e9)}")

            case 'cron':
                if 'schedule' not in t_rules:
                    raise ValueError(f"cron trigger type is specified but cron schedule not found !")

                self._scheduler.schedule_event(t_rules['schedule'], 'time_event')

            case 'quote': 
                self._trig_on_quote = True
                raise ValueError(f"quote trigger NOT IMPLEMENTED YET")

            case 'trade': 
                self._trig_on_trade = True
                raise ValueError(f"trade trigger NOT IMPLEMENTED YET")

            case 'orderbook' | 'ob': 
                self._trig_on_book = True
                raise ValueError(f"orderbook trigger NOT IMPLEMENTED YET")

            case _: 
                raise ValueError(f"Wrong trigger type {t_rules['type']}")

        # - check fit spec - - - - -
        _last_fit_data_can_be_used = pd.Timestamp(self.time())
        match f_rules.get('type'):
            case 'cron':
                if 'schedule' not in f_rules:
                    raise ValueError(f"cron fit trigger type is specified but cron schedule not found !")

                self._scheduler.schedule_event(f_rules['schedule'], 'fit_event')
                _last_fit_data_can_be_used = self._scheduler.get_event_last_time('fit_event')

            case 'bar':
                raise ValueError("Raw 'bar' type is not supported for fitting spec yet, please use cron type !")

            case _: 
                # if schedule is not specified just do not arm the task 
                # only initial fit will be called
                pass

        # - we can't just call on_fit right now because not all market data may be ready
        # - so we just mark it as not called yet
        self.__init_fit_was_called = False
        self.__init_fit_args = (None, _last_fit_data_can_be_used) 
   
    def _process_incoming_data(self, channel: CtrlChannel):
        _fails_counter = 0 
        logger.info("(StrategyContext) Start processing market data")

        while channel.control.is_set():
            # - start loop latency measurement 
            _SW.start('StrategyContext._process_incoming_data')

            # - waiting for incoming market data
            symbol, d_type, data = channel.queue.get()

            # - process data if handler is registered
            handler = self.__handlers.get(d_type)
            _SW.start('StrategyContext.handler')
            _strategy_trigger_on_event = handler(self, symbol, data) if handler else None
            _SW.stop('StrategyContext.handler')

            # - check if it still didn't call on_fit() for first time
            if not self.__init_fit_was_called:
                self._processing_fit_event(None, self.__init_fit_args)

            if _strategy_trigger_on_event:

                # - if fit was not called - skip on_event call
                if not self.__init_fit_was_called:
                    _SW.stop('StrategyContext._process_incoming_data')
                    logger.warning(f"[{self.time()}] {self.strategy.__class__.__name__}::on_event() is SKIPPED for now because on_fit() was not called yet !")
                    continue

                # - if strategy still fitting - skip on_event call
                if self.__fit_is_running:
                    _SW.stop('StrategyContext._process_incoming_data')
                    logger.warning(f"[{self.time()}] {self.strategy.__class__.__name__}::on_event() is SKIPPED for now because is being still fitting !")
                    continue

                try:
                    _SW.start('strategy.on_event')
                    self.strategy.on_event(self, _strategy_trigger_on_event)
                    _fails_counter = 0
                except Exception as strat_error:
                    # - probably we need some cooldown interval after exception to prevent flooding
                    logger.error(f"[{self.time()}]: Strategy {self.strategy.__class__.__name__} raised an exception: {strat_error}")
                    logger.opt(colors=False).error(traceback.format_exc())

                    #  - we stop execution after let's say maximal number of errors in a row
                    if (_fails_counter := _fails_counter + 1) >= StrategyContext.MAX_NUMBER_OF_STRATEGY_FAILURES:
                        logger.error("STRATEGY FAILURES IN THE ROW EXCEEDED MAX ALLOWED NUMBER - STOPPING ...")
                        channel.stop()
                        break
                finally:
                    _SW.stop('strategy.on_event')

            # - stop loop latency measurement 
            _SW.stop('StrategyContext._process_incoming_data')

            # - notify poition and portfolio loggers
            self._logging.notify(self.time())

        _SW.stop('StrategyContext._process_incoming_data')
        logger.info("(StrategyContext) Market data processing stopped")

    def _invoke_on_fit(self, current_fit_time: str | pd.Timestamp, prev_fit_time: str | pd.Timestamp | None):
        try:
            self.__fit_is_running = True
            logger.debug(f"[{self.time()}]: Invoking {self.strategy.__class__.__name__} on_fit('{current_fit_time}', '{prev_fit_time}')")
            _SW.start('strategy.on_fit')
            self.strategy.on_fit(self, current_fit_time, prev_fit_time)
            logger.debug(f"[{self.time()}]: {self.strategy.__class__.__name__} is fitted")
        except Exception as strat_error:
            logger.error(f"[{self.time()}]: Strategy {self.strategy.__class__.__name__} on_fit('{current_fit_time}', '{prev_fit_time}') raised an exception: {strat_error}")
            logger.opt(colors=False).error(traceback.format_exc())
        finally:
            self.__fit_is_running = False
            self.__init_fit_was_called = True
            _SW.stop('strategy.on_fit')
        return None

    def _processing_time_event(self, symbol: str, data: Any) -> TriggerEvent | None:
        """
        When scheduled time event is happened - we need to invoke strategy on_event method
        """
        return TriggerEvent(self.time(), 'time', None, data)

    def _processing_fit_event(self, symbol: str | None, data: Any) -> TriggerEvent | None:
        """
        When scheduled fit event is happened - we need to invoke strategy on_fit method
        """
        if not self._cache.is_data_ready():
            return None

        # times are in seconds here
        prev_fit_time, now_fit_time = data

        # - we need to run this in separate thread
        self.__fit_is_running = True
        self._run_in_thread_pool(self._invoke_on_fit, (
            pd.Timestamp(now_fit_time, unit='s'), 
            pd.Timestamp(prev_fit_time, unit='s') if prev_fit_time else None
        ))

        return None

    def _processing_hist_bar(self, symbol: str, bar: Bar) -> TriggerEvent | None:
        # - processing single historical bar
        #   here it just updates cache - historical bar can't trigger strategy logic
        self._cache.update_by_bar(symbol, bar)
        return None

    def _processing_hist_bars(self, symbol: str, bars: List[Bar]) -> TriggerEvent | None:
        # - processing historical bars as list
        for b in bars:
            self._processing_hist_bar(symbol, b)
        return None

    @_SW.watch('StrategyContext')
    def _processing_bar(self, symbol: str, bar: Bar) -> TriggerEvent | None:
        # - processing current bar's update
        self._cache.update_by_bar(symbol, bar)

        # - check if it's time to trigger the on_event if it's configured
        if self._trig_on_bar:
            t = self.exchange_service.time().item()
            _time_to_trigger = t % self._trig_bar_freq_nsec >= self._trig_interval_in_bar_nsec
            if _time_to_trigger: 
                # we want to trigger only first one - not every
                if not self._current_bar_trigger_processed:
                    self._current_bar_trigger_processed = True
                    return TriggerEvent(self.time(), 'bar', self._symb_to_instr.get(symbol), bar)
            else:
                self._current_bar_trigger_processed = False
        return None

    def _processing_trade(self, symbol: str, trade: Trade) -> TriggerEvent | None:
        if self._trig_on_trade:
            return TriggerEvent(self.time(), 'trade', symbol, trade)
        return None

    def _processing_quote(self, symbol: str, quote: Quote) -> TriggerEvent | None:
        # - TODO: here we can apply throttlings or filters
        #  - let's say we can skip quotes if bid & ask is not changed
        #  - or we can collect let's say N quotes before sending to strategy
        if self._trig_on_quote:
            return TriggerEvent(self.time(), 'quote', symbol, quote)
        return None

    @_SW.watch('StrategyContext')
    def _processing_order(self, symbol: str, order: Order) -> TriggerEvent | None:
        logger.debug(f"[{order.id} / {order.client_id}] : {order.type} {order.side} {order.quantity} of {symbol} { (' @ ' + str(order.price)) if order.price else '' } -> [{order.status}]")
        # - check if we want to trigger any strat's logic on order
        return None

    @_SW.watch('StrategyContext')
    def _processing_deals(self, symbol: str, deals: List[Deal]) -> TriggerEvent | None:
        # - log deals in storage
        self._logging.save_deals(symbol, deals)
        for d in deals:
            logger.debug(f"Executed {d.amount} @ {d.price} of {symbol} for order {d.order_id}")
        return None

    def ohlc(self, instrument: str | Instrument, timeframe: str) -> OHLCV:
        return self._cache.get_ohlcv(instrument if isinstance(instrument, str) else instrument.symbol, timeframe)

    def _create_and_update_positions(self):
        for instrument in self.instruments:
            symb = instrument.symbol
            self.positions[symb] = self.exchange_service.get_position(instrument)
            # - check if we need any aux instrument for calculating pnl ?
            aux = lookup.find_aux_instrument_for(instrument, self.exchange_service.get_base_currency())
            if aux is not None:
                instrument._aux_instrument = aux
                self.instruments.append(aux)
                self.positions[aux.symbol] = self.exchange_service.get_position(aux)

    def start(self):
        if self._is_initilized :
            raise ValueError("Strategy is already started !")

        # - create positions for instruments
        self._create_and_update_positions()

        # - get actual positions from exchange
        _symbols = []
        for instr in self.instruments:
            # process instruments - need to find convertors etc
            self._cache.init_ohlcv(instr.symbol)
            _symbols.append(instr.symbol)

        # - create incoming market data processing
        # self._thread_data_loop = AsyncioThreadRunner(self.data_provider.get_communication_channel())
        self._thread_data_loop = Thread(
            target=self._process_incoming_data, 
            args=(self.data_provider.get_communication_channel(),), 
            daemon=True
        )
        # self._thread_data_loop.add(self._process_incoming_market_data)

        # - subscribe to market data
        logger.info(f"(StrategyContext) Subscribing to {self._market_data_subcription_type} updates using {self._market_data_subcription_params} for \n\t{_symbols} ")
        self.data_provider.subscribe(self._market_data_subcription_type, _symbols, **self._market_data_subcription_params)

        # - initialize strategy loggers
        self._logging.initialize(
            self.time(), self.positions, self.exchange_service.get_account().get_balances()
        )

        # - initialize strategy (should we do that after any first market data received ?)
        if not self._is_initilized:
            try:
                self.strategy.on_start(self)
                self._is_initilized = True
            except Exception as strat_error:
                logger.error(f"(StrategyContext) Strategy {self.strategy.__class__.__name__} raised an exception in on_start: {strat_error}")
                logger.error(traceback.format_exc())
                return

        self._thread_data_loop.start()
        logger.info("(StrategyContext) strategy is started")

    def stop(self):
        if self._thread_data_loop:
            self.data_provider.get_communication_channel().stop()
            self._thread_data_loop.join()
            try:
                self.strategy.on_stop(self)
            except Exception as strat_error:
                logger.error(f"(StrategyContext) Strategy {self.strategy.__class__.__name__} raised an exception in on_stop: {strat_error}")
                logger.error(traceback.format_exc())
            self._thread_data_loop = None

        # - close logging
        self._logging.close()
        self.get_latencies_report()

    def get_latencies_report(self):
        for l in _SW.latencies.keys():
            logger.info(f"\t<w>{l}</w> took <r>{_SW.latency_sec(l):.7f}</r> secs")

    def populate_parameters_to_strategy(self, strategy: IStrategy, **kwargs):
        _log_info = ""
        for k,v in kwargs.items():
            if k.startswith('_'):
                raise ValueError("Internal variable can't be set from external parameter !")
            if hasattr(strategy, k):
                strategy.__dict__[k] = v
                _log_info += f"\n\tset <green>{k}</green> <- <red>{v}</red>"
        if _log_info:
            logger.info("(StrategyContext) set strategy parameters:" + _log_info)

    def time(self) -> dt_64:
        return self.exchange_service.time()

    def _generate_order_client_id(self, symbol: str) -> str:
        self.__order_id += 1
        return self.__strategy_id + symbol + '_' + str(self.__order_id)

    @_SW.watch('StrategyContext')
    def trade(self, instr_or_symbol: Instrument | str, amount:float, price: float|None=None, time_in_force='gtc') -> Order:
        _SW.start('send_order')
        instrument: Instrument | None = self._symb_to_instr.get(instr_or_symbol) if isinstance(instr_or_symbol, str) else instr_or_symbol
        if instrument is None:
            raise ValueError(f"Can't find instrument for symbol {instr_or_symbol}")

        # - adjust size
        size_adj = _round_down_at_min_qty(abs(amount), instrument.min_size)
        if size_adj == 0:
            raise ValueError(f"Attempt to trade size {abs(amount)} less than minimal allowed {instrument.min_size} !")

        side = 'buy' if amount > 0 else 'sell'
        type = 'limit' if price is not None else 'market'
        logger.info(f"(StrategyContext) sending {type} {side} for {size_adj} of {instrument.symbol} ...")
        client_id = self._generate_order_client_id(instrument.symbol)
        order = self.exchange_service.send_order(instrument, side, type, size_adj, price, time_in_force=time_in_force, client_id=client_id) 

        return order

    @_SW.watch('StrategyContext')
    def cancel(self, instr_or_symbol: Instrument | str):
        instrument: Instrument | None = self._symb_to_instr.get(instr_or_symbol) if isinstance(instr_or_symbol, str) else instr_or_symbol
        if instrument is None:
            raise ValueError(f"Can't find instrument for symbol {instr_or_symbol}")
        for o in self.exchange_service.get_orders(instrument.symbol):
            self.exchange_service.cancel_order(o.id)

    def quote(self, symbol: str) -> Quote | None:
        return self.data_provider.get_quote(symbol)

    def get_capital(self) -> float:
        return self.exchange_service.get_capital()

    def get_reserved(self, instrument: Instrument) -> float:
        return self.exchange_service.get_account().get_reserved(instrument)

    @_SW.watch('StrategyContext')
    def get_historical_ohlcs(self, instrument: Instrument | str, timeframe: str, length: int) -> OHLCV | None:
        """
        Helper for historical ohlc data
        """
        instr = self._symb_to_instr.get(instrument) if isinstance(instrument, str) else instrument

        if instr is None:
            logger.warning(f"Can't find instrument for {instrument} symbol !")
            return None

        # - first check if we can use cached series
        rc = self.ohlc(instr, timeframe)
        if len(rc) >= length:
            return rc

        # - send request for historical data
        bars = self.data_provider.get_historical_ohlcs(instr.symbol, timeframe, length) 
        r = self._cache.update_by_bars(instr.symbol, timeframe, bars)
        return r 

    def _run_in_thread_pool(self, func: Callable, args=()):
        """
        For backtester we need to override this method and just call function
        """
        if self.__pool is None:
            self.__pool = ThreadPool(2)
        self.__pool.apply_async(func, args)


if __name__ == '__main__':
    from qubx.utils.runner import create_strategy_context
    import time, sys
    filename, accounts, paths = '../experiments/configs/zero_test_scheduler.yaml', "../experiments/configs/.env", ['../']

    # - create context
    ctx = create_strategy_context(filename, accounts, paths)
    if ctx is None:
        sys.exit(0)

    # - run main loop
    try:
        ctx.start()

        # - just wake up every 60 sec and check if it's OK 
        while True: 
            time.sleep(60)

    except KeyboardInterrupt:
        ctx.stop()
        time.sleep(1)
        sys.exit(0)


from typing import Any, Dict, List, Optional, Tuple
import pandas as pd

from qubx import logger
from qubx.core.basics import Order, Deal, Position


EXCHANGE_ALIASES = {
    "binance.um": "binanceusdm",
    "binance.cm": "binancecoinm",
    "kraken.f": "krakenfutures"
}

DATA_PROVIDERS_ALIASES = EXCHANGE_ALIASES | { "binance": "binanceqv" }


def ccxt_convert_order_info(symbol: str, raw: Dict[str,Any]) -> Order:
    """
    Convert CCXT excution record to Order object
    """
    ri = raw['info']
    amnt = float(ri.get('origQty', raw.get('amount')))
    price = raw['price']
    status = raw['status']
    side = raw['side'].upper()
    _type = ri.get('type', raw.get('type')).upper()
    if status == 'open':
        status = ri.get('status', status)  # for filled / part_filled ?

    return Order(
        id=raw['id'],
        type=_type,
        symbol=symbol,
        time = pd.Timestamp(raw['timestamp'], unit='ms'), # type: ignore
        quantity = amnt, 
        price= float(price) if price is not None else 0.0,  
        side = side,
        status = status.upper(),
        time_in_force = raw['timeInForce'],
        client_id = raw['clientOrderId'],
        cost = float(raw['cost']),
    )


def ccxt_convert_deal_info(raw: Dict[str,Any]) -> Deal:
    fee_amount = None
    fee_currency = None
    if 'fee' in raw:
        fee_amount = float(raw['fee']['cost'])
        fee_currency = raw['fee']['currency']
    return Deal(
        id=raw['id'],
        order_id=raw['order'],
        time = pd.Timestamp(raw['timestamp'], unit='ms'), # type: ignore
        amount=float(raw['amount']) * (-1 if raw['side'] == 'sell' else +1),
        price=float(raw['price']),
        aggressive=raw['takerOrMaker'] == 'taker',
        fee_amount=fee_amount,
        fee_currency=fee_currency,
    )


def ccxt_extract_deals_from_exec(report: Dict[str,Any]) -> List[Deal]:
    """
    Small helper for extracting deals (trades) from CCXT execution report
    """
    deals = list()
    if (trades := report.get('trades')):
        for t in trades:
            deals.append(ccxt_convert_deal_info(t))
    return deals


def ccxt_restore_position_from_deals(
    pos: Position, current_volume: float, deals: List[Deal], reserved_amount:float=0.0
) -> Position:
    if  current_volume != 0:
        instr = pos.instrument
        _last_deals = []

        # - try to find last deals that led to this position
        for d in sorted(deals, key=lambda x: x.time, reverse=True):
            current_volume -= d.amount
            # - spot case when fees may be deducted from the base coin
            #   that may decrease total amount
            if d.fee_amount is not None:
                if instr.base == d.fee_currency:
                    current_volume += d.fee_amount
            # print(d.amount, current_volume)
            _last_deals.insert(0, d)

            # - take in account reserves
            if abs(current_volume) - abs(reserved_amount) < instr.min_size_step:
                break
            
        # - reset to 0
        pos.reset()

        if abs(current_volume) - abs(reserved_amount) > instr.min_size_step:
            # - - - TODO - - - !!!!
            logger.warning(f"Couldn't restore full deals history for {instr.symbol} symbol. Qubx will use zero position !")
        else:
            for d in _last_deals:
                pos.update_position_by_deal(d)
                if d.fee_amount is not None:
                    if instr.base == d.fee_currency:
                        pos.quantity -= d.fee_amount
    return pos


"""
   Misc graphics handy utilitites to be used in interactive analysis
"""
import numpy as np
import pandas as pd
import statsmodels.tsa.stattools as st
import matplotlib
import matplotlib.pyplot as plt
import colorsys
import datetime

import matplotlib.colors as mc
import matplotlib.ticker as mticker
from matplotlib import colors as mcolors, pyplot as plt
from matplotlib.collections import LineCollection, PolyCollection
from matplotlib.dates import num2date, date2num
from matplotlib.lines import TICKLEFT, TICKRIGHT, Line2D
from matplotlib.patches import Rectangle
from matplotlib.transforms import Affine2D
from cycler import cycler

from qubx.utils.time import infer_series_frequency


DARK_MATLPLOT_THEME = [
    ('backend', 'module://matplotlib_inline.backend_inline'),
    ('interactive', True),
    ('lines.color', '#5050f0'),
    ('text.color', '#d0d0d0'),
    ('axes.facecolor', '#000000'),
    ('axes.edgecolor', '#404040'),
    ('axes.grid', True),
    ('axes.labelsize', 'large'),
    ('axes.labelcolor', 'green'),
    ('axes.prop_cycle', cycler('color', ['#08F7FE', '#00ff41', '#FE53BB', '#F5D300', '#449AcD', 'g',
                                         '#f62841', 'y', '#088487', '#E24A33', '#f01010'])),
    ('legend.fontsize', 'small'),
    ('legend.fancybox', False),
    ('legend.edgecolor', '#305030'),
    ('legend.shadow', False),
    ('lines.antialiased', True),
    ('lines.linewidth', 0.8),  # reduced line width
    ('patch.linewidth', 0.5),
    ('patch.antialiased', True),
    ('xtick.color', '#909090'),
    ('ytick.color', '#909090'),
    ('xtick.labelsize', 'large'),
    ('ytick.labelsize', 'large'),
    ('grid.color', '#404040'),
    ('grid.linestyle', '--'),
    ('grid.linewidth', 0.5),
    ('grid.alpha', 0.8),
    ('figure.figsize', [12.0, 5.0]),
    ('figure.dpi', 80.0),
    ('figure.facecolor', '#050505'),
    ('figure.edgecolor', (1, 1, 1, 0)),
    ('figure.subplot.bottom', 0.125),
    ('savefig.facecolor', '#000000'),
]

LIGHT_MATPLOT_THEME = [
    ('backend', 'module://matplotlib_inline.backend_inline'),
    ('interactive', True),
    ('lines.color', '#101010'),
    ('text.color', '#303030'),
    ('lines.antialiased', True),
    ('lines.linewidth', 1),
    ('patch.linewidth', 0.5),
    ('patch.facecolor', '#348ABD'),
    ('patch.edgecolor', '#eeeeee'),
    ('patch.antialiased', True),
    ('axes.facecolor', '#fafafa'),
    ('axes.edgecolor', '#d0d0d0'),
    ('axes.linewidth', 1),
    ('axes.titlesize', 'x-large'),
    ('axes.labelsize', 'large'),
    ('axes.labelcolor', '#555555'),
    ('axes.axisbelow', True),
    ('axes.grid', True),
    ('axes.prop_cycle', cycler('color', ['#6792E0', '#27ae60', '#c44e52', '#975CC3', '#ff914d', '#77BEDB',
                                         '#303030', '#4168B7', '#93B851', '#e74c3c', '#bc89e0', '#ff711a',
                                         '#3498db', '#6C7A89'])),
    ('legend.fontsize', 'small'),
    ('legend.fancybox', False),
    ('xtick.color', '#707070'),
    ('ytick.color', '#707070'),
    ('grid.color', '#606060'),
    ('grid.linestyle', '--'),
    ('grid.linewidth', 0.5),
    ('grid.alpha', 0.3),
    ('figure.figsize', [8.0, 5.0]),
    ('figure.dpi', 80.0),
    ('figure.facecolor', '#ffffff'),
    ('figure.edgecolor', '#ffffff'),
    ('figure.subplot.bottom', 0.1)
]


def fig(w=16, h=5, dpi=96, facecolor=None, edgecolor=None, num=None):
    """
    Simple helper for creating figure
    """
    return plt.figure(num=num, figsize=(w, h), dpi=dpi, facecolor=facecolor, edgecolor=edgecolor)


def subplot(shape, loc, rowspan=2, colspan=1):
    """
    Some handy grid splitting for plots. Example for 2x2:
    
    >>> subplot(22, 1); plt.plot([-1,2,-3])
    >>> subplot(22, 2); plt.plot([1,2,3])
    >>> subplot(22, 3); plt.plot([1,2,3])
    >>> subplot(22, 4); plt.plot([3,-2,1])

    same as following

    >>> subplot((2,2), (0,0)); plt.plot([-1,2,-3])
    >>> subplot((2,2), (0,1)); plt.plot([1,2,3])
    >>> subplot((2,2), (1,0)); plt.plot([1,2,3])
    >>> subplot((2,2), (1,1)); plt.plot([3,-2,1])

    :param shape: scalar (like matlab subplot) or tuple
    :param loc: scalar (like matlab subplot) or tuple
    :param rowspan: rows spanned
    :param colspan: columns spanned
    """
    isscalar = lambda x: not isinstance(x, (list, tuple, dict, np.ndarray))

    if isscalar(shape):
        if 0 < shape < 100:
            shape = (max(shape // 10, 1), max(shape % 10, 1))
        else:
            raise ValueError("Wrong scalar value for shape. It should be in range (1...99)")

    if isscalar(loc):
        nm = max(shape[0], 1) * max(shape[1], 1)
        if 0 < loc <= nm:
            x = (loc - 1) // shape[1]
            y = loc - 1 - shape[1] * x
            loc = (x, y)
        else:
            raise ValueError("Wrong scalar value for location. It should be in range (1...%d)" % nm)

    return plt.subplot2grid(shape, loc=loc, rowspan=rowspan, colspan=colspan)


def sbp(shape, loc, r=1, c=1):
    """
    Just shortcut for subplot(...) function

    :param shape: scalar (like matlab subplot) or tuple
    :param loc: scalar (like matlab subplot) or tuple
    :param r: rows spanned
    :param c: columns spanned
    :return:
    """
    return subplot(shape, loc, rowspan=r, colspan=c)


def vline(ax, x, c, lw=1, ls='--'):
    x = pd.to_datetime(x) if isinstance(x, str) else x
    if not isinstance(ax, (list, tuple)):
        ax = [ax]
    for a in ax:
        a.axvline(x, 0, 1, c=c, lw=1, linestyle=ls)


def hline(*zs, mirror=True):
    [plt.axhline(z, ls='--', c='r', lw=0.5) for z in zs]
    if mirror:
        [plt.axhline(-z, ls='--', c='r', lw=0.5) for z in zs]


def ellips(ax, x, y, c='r', r=2.5, lw=2, ls='-'):
    """
    Draw ellips annotation on specified plot at (x,y) point
    """
    from matplotlib.patches import Ellipse
    x = pd.to_datetime(x) if isinstance(x, str) else x
    w, h = (r, r) if np.isscalar(r) else (r[0], r[1])
    ax.add_artist(Ellipse(xy=[x, y], width=w, height=h, angle=0, fill=False, color=c, lw=lw, ls=ls))


def set_mpl_theme(theme: str):
    import plotly.io as pio

    if 'dark' in theme.lower():
        pio.templates.default = "plotly_dark"
        for (k, v) in DARK_MATLPLOT_THEME:
            matplotlib.rcParams[k] = v

    elif 'light' in theme.lower():
        pio.templates.default = "plotly_white"
        for (k, v) in LIGHT_MATPLOT_THEME:
            matplotlib.rcParams[k] = v


def adjust_lightness(color, amount=0.5):
    try:
        c = mc.cnames[color]
    except:
        c = color
    c = colorsys.rgb_to_hls(*mc.to_rgb(c))
    return colorsys.hls_to_rgb(c[0], max(0, min(1, amount * c[1])), c[2])


def plot_day_summary_oclh(ax, quotes, ticksize=3, colorup='k', colordown='r'):
    """Plots day summary

        Represent the time, open, close, high, low as a vertical line
        ranging from low to high.  The left tick is the open and the right
        tick is the close.

    Parameters
    ----------
    ax : `Axes`
        an `Axes` instance to plot to
    quotes : sequence of (time, open, close, high, low, ...) sequences
        data to plot.  time must be in float date format - see date2num
    ticksize : int
        open/close tick marker in points
    colorup : color
        the color of the lines where close >= open
    colordown : color
        the color of the lines where close <  open

    Returns
    -------
    lines : list
        list of tuples of the lines added (one tuple per quote)
    """
    return _plot_day_summary(ax, quotes, ticksize=ticksize,
                             colorup=colorup, colordown=colordown,
                             ochl=True)


def plot_day_summary_ohlc(ax, quotes, ticksize=3, colorup='k', colordown='r'):
    """Plots day summary

        Represent the time, open, high, low, close as a vertical line
        ranging from low to high.  The left tick is the open and the right
        tick is the close.

    Parameters
    ----------
    ax : `Axes`
        an `Axes` instance to plot to
    quotes : sequence of (time, open, high, low, close, ...) sequences
        data to plot.  time must be in float date format - see date2num
    ticksize : int
        open/close tick marker in points
    colorup : color
        the color of the lines where close >= open
    colordown : color
        the color of the lines where close <  open

    Returns
    -------
    lines : list
        list of tuples of the lines added (one tuple per quote)
    """
    return _plot_day_summary(ax, quotes, ticksize=ticksize,
                             colorup=colorup, colordown=colordown,
                             ochl=False)


def _plot_day_summary(ax, quotes, ticksize=3, colorup='k', colordown='r', ochl=True):
    """Plots day summary


        Represent the time, open, high, low, close as a vertical line
        ranging from low to high.  The left tick is the open and the right
        tick is the close.

    Parameters
    ----------
    ax : `Axes`
        an `Axes` instance to plot to
    quotes : sequence of quote sequences
        data to plot.  time must be in float date format - see date2num
        (time, open, high, low, close, ...) vs
        (time, open, close, high, low, ...)
        set by `ochl`
    ticksize : int
        open/close tick marker in points
    colorup : color
        the color of the lines where close >= open
    colordown : color
        the color of the lines where close <  open
    ochl: bool
        argument to select between ochl and ohlc ordering of quotes

    Returns
    -------
    lines : list
        list of tuples of the lines added (one tuple per quote)
    """
    # unfortunately this has a different return type than plot_day_summary2_*
    lines = []
    for q in quotes:
        if ochl:
            t, open, close, high, low = q[:5]
        else:
            t, open, high, low, close = q[:5]

        if close >= open:
            color = colorup
        else:
            color = colordown

        vline = Line2D(xdata=(t, t), ydata=(low, high),
                       color=color,
                       antialiased=False,  # no need to antialias vert lines
                       )

        oline = Line2D(xdata=(t, t), ydata=(open, open),
                       color=color,
                       antialiased=False,
                       marker=TICKLEFT,
                       markersize=ticksize,
                       )

        cline = Line2D(xdata=(t, t), ydata=(close, close),
                       color=color,
                       antialiased=False,
                       markersize=ticksize,
                       marker=TICKRIGHT)

        lines.extend((vline, oline, cline))
        ax.add_line(vline)
        ax.add_line(oline)
        ax.add_line(cline)

    ax.autoscale_view()

    return lines


def candlestick_ochl(ax, quotes, width=0.2, colorup='k', colordown='r', alpha=1.0):
    """
    Plot the time, open, close, high, low as a vertical line ranging
    from low to high.  Use a rectangular bar to represent the
    open-close span.  If close >= open, use colorup to color the bar,
    otherwise use colordown

    Parameters
    ----------
    ax : `Axes`
        an Axes instance to plot to
    quotes : sequence of (time, open, close, high, low, ...) sequences
        As long as the first 5 elements are these values,
        the record can be as long as you want (e.g., it may store volume).

        time must be in float days format - see date2num

    width : float
        fraction of a day for the rectangle width
    colorup : color
        the color of the rectangle where close >= open
    colordown : color
         the color of the rectangle where close <  open
    alpha : float
        the rectangle alpha level

    Returns
    -------
    ret : tuple
        returns (lines, patches) where lines is a list of lines
        added and patches is a list of the rectangle patches added

    """
    return _candlestick(ax, quotes, width=width, colorup=colorup,
                        colordown=colordown,
                        alpha=alpha, ochl=True)


def candlestick_ohlc(ax, quotes, width=0.2, colorup='k', colordown='r', alpha=1.0):
    """
    Plot the time, open, high, low, close as a vertical line ranging
    from low to high.  Use a rectangular bar to represent the
    open-close span.  If close >= open, use colorup to color the bar,
    otherwise use colordown

    Parameters
    ----------
    ax : `Axes`
        an Axes instance to plot to
    quotes : sequence of (time, open, high, low, close, ...) sequences
        As long as the first 5 elements are these values,
        the record can be as long as you want (e.g., it may store volume).

        time must be in float days format - see date2num

    width : float
        fraction of a day for the rectangle width
    colorup : color
        the color of the rectangle where close >= open
    colordown : color
         the color of the rectangle where close <  open
    alpha : float
        the rectangle alpha level

    Returns
    -------
    ret : tuple
        returns (lines, patches) where lines is a list of lines
        added and patches is a list of the rectangle patches added

    """
    return _candlestick(ax, quotes, width=width, colorup=colorup,
                        colordown=colordown,
                        alpha=alpha, ochl=False)


def _candlestick(ax, quotes, width=0.2, colorup='k', colordown='r', alpha=1.0, ochl=True):
    """
    Plot the time, open, high, low, close as a vertical line ranging
    from low to high.  Use a rectangular bar to represent the
    open-close span.  If close >= open, use colorup to color the bar,
    otherwise use colordown

    Parameters
    ----------
    ax : `Axes`
        an Axes instance to plot to
    quotes : sequence of quote sequences
        data to plot.  time must be in float date format - see date2num
        (time, open, high, low, close, ...) vs
        (time, open, close, high, low, ...)
        set by `ochl`
    width : float
        fraction of a day for the rectangle width
    colorup : color
        the color of the rectangle where close >= open
    colordown : color
         the color of the rectangle where close <  open
    alpha : float
        the rectangle alpha level
    ochl: bool
        argument to select between ochl and ohlc ordering of quotes

    Returns
    -------
    ret : tuple
        returns (lines, patches) where lines is a list of lines
        added and patches is a list of the rectangle patches added

    """

    OFFSET = width / 2.0

    lines = []
    patches = []
    for q in quotes:
        if ochl:
            t, open, close, high, low = q[:5]
        else:
            t, open, high, low, close = q[:5]

        if close >= open:
            color = colorup
            edgecolor = adjust_lightness(color, 1.5)
            lower = open
            height = close - open
        else:
            color = colordown
            edgecolor = adjust_lightness(color, 1.2)
            lower = close
            height = open - close

        vline = Line2D(
            xdata=(t, t), ydata=(low, high),
            # color=adjust_lightness(color, 1.3),
            color=color,
            linewidth=1,
            antialiased=True,
        )

        rect = Rectangle(
            xy=(t - OFFSET, lower),
            width=width,
            height=height,
            facecolor=color,
            # edgecolor=color,
            # facecolor=adjust_lightness(color, 1.3),
            edgecolor=edgecolor,
            lw=0.75

        )
        rect.set_alpha(alpha)

        lines.append(vline)
        patches.append(rect)
        ax.add_line(vline)
        ax.add_patch(rect)
    ax.autoscale_view()

    return lines, patches


def _check_input(opens, closes, highs, lows, miss=-1):
    """Checks that *opens*, *highs*, *lows* and *closes* have the same length.
    NOTE: this code assumes if any value open, high, low, close is
    missing (*-1*) they all are missing

    Parameters
    ----------
    ax : `Axes` an Axes instance to plot to
    opens : sequence of opening values
    highs : sequence of high values
    lows : sequence of low values
    closes : sequence of closing values
    miss : identifier of the missing data

    Raises
    ------
    ValueError
        if the input sequences don't have the same length
    """

    def _missing(sequence, miss=-1):
        """Returns the index in *sequence* of the missing data, identified by
        *miss*

        Parameters
        ----------
        sequence :
            sequence to evaluate
        miss :
            identifier of the missing data

        Returns
        -------
        where_miss: numpy.ndarray
            indices of the missing data
        """
        return np.where(np.array(sequence) == miss)[0]

    same_length = len(opens) == len(highs) == len(lows) == len(closes)
    _missopens = _missing(opens)
    same_missing = ((_missopens == _missing(highs)).all() and
                    (_missopens == _missing(lows)).all() and
                    (_missopens == _missing(closes)).all())

    if not (same_length and same_missing):
        msg = ("*opens*, *highs*, *lows* and *closes* must have the same"
               " length. NOTE: this code assumes if any value open, high,"
               " low, close is missing (*-1*) they all must be missing.")
        raise ValueError(msg)


def plot_day_summary2_ochl(ax, opens, closes, highs, lows, ticksize=4,
                           colorup='k', colordown='r'):
    """Represent the time, open, close, high, low,  as a vertical line
    ranging from low to high.  The left tick is the open and the right
    tick is the close.

    Parameters
    ----------
    ax : `Axes`
        an Axes instance to plot to
    opens : sequence of opening values
    closes : sequence of closing values
    highs : sequence of high values
    lows : sequence of low values
    ticksize : int
        size of open and close ticks in points
    colorup : color
        the color of the lines where close >= open
    colordown : color
         the color of the lines where close <  open

    Returns
    -------
    ret : list
        a list of lines added to the axes
    """

    return plot_day_summary2_ohlc(ax, opens, highs, lows, closes, ticksize,
                                  colorup, colordown)


def plot_day_summary2_ohlc(ax, opens, highs, lows, closes, ticksize=4,
                           colorup='k', colordown='r'):
    """Represent the time, open, high, low, close as a vertical line
    ranging from low to high.  The left tick is the open and the right
    tick is the close.
    *opens*, *highs*, *lows* and *closes* must have the same length.
    NOTE: this code assumes if any value open, high, low, close is
    missing (*-1*) they all are missing

    Parameters
    ----------
    ax : `Axes`
        an Axes instance to plot to
    opens : sequence
        sequence of opening values
    highs : sequence
        sequence of high values
    lows : sequence
        sequence of low values
    closes : sequence
        sequence of closing values
    ticksize : int
        size of open and close ticks in points
    colorup : color
        the color of the lines where close >= open
    colordown : color
         the color of the lines where close <  open

    Returns
    -------
    ret : list
        a list of lines added to the axes
    """

    _check_input(opens, highs, lows, closes)

    rangeSegments = [((i, low), (i, high)) for i, low, high in
                     zip(range(len(lows)), lows, highs) if low != -1]

    # the ticks will be from ticksize to 0 in points at the origin and
    # we'll translate these to the i, close location
    openSegments = [((-ticksize, 0), (0, 0))]

    # the ticks will be from 0 to ticksize in points at the origin and
    # we'll translate these to the i, close location
    closeSegments = [((0, 0), (ticksize, 0))]

    offsetsOpen = [(i, open) for i, open in
                   zip(range(len(opens)), opens) if open != -1]

    offsetsClose = [(i, close) for i, close in
                    zip(range(len(closes)), closes) if close != -1]

    scale = ax.figure.dpi * (1.0 / 72.0)

    tickTransform = Affine2D().scale(scale, 0.0)

    colorup = mcolors.to_rgba(colorup)
    colordown = mcolors.to_rgba(colordown)
    colord = {True: colorup, False: colordown}
    colors = [colord[open < close] for open, close in
              zip(opens, closes) if open != -1 and close != -1]

    useAA = 0,  # use tuple here
    lw = 1,  # and here
    rangeCollection = LineCollection(rangeSegments,
                                     colors=colors,
                                     linewidths=lw,
                                     antialiaseds=useAA,
                                     )

    openCollection = LineCollection(openSegments,
                                    colors=colors,
                                    antialiaseds=useAA,
                                    linewidths=lw,
                                    offsets=offsetsOpen,
                                    transOffset=ax.transData,
                                    )
    openCollection.set_transform(tickTransform)

    closeCollection = LineCollection(closeSegments,
                                     colors=colors,
                                     antialiaseds=useAA,
                                     linewidths=lw,
                                     offsets=offsetsClose,
                                     transOffset=ax.transData,
                                     )
    closeCollection.set_transform(tickTransform)

    minpy, maxx = (0, len(rangeSegments))
    miny = min([low for low in lows if low != -1])
    maxy = max([high for high in highs if high != -1])
    corners = (minpy, miny), (maxx, maxy)
    ax.update_datalim(corners)
    ax.autoscale_view()

    # add these last
    ax.add_collection(rangeCollection)
    ax.add_collection(openCollection)
    ax.add_collection(closeCollection)
    return rangeCollection, openCollection, closeCollection


def candlestick2_ochl(ax, opens, closes, highs, lows, width=4,
                      colorup='k', colordown='r',
                      alpha=0.75):
    """Represent the open, close as a bar line and high low range as a
    vertical line.

    Preserves the original argument order.


    Parameters
    ----------
    ax : `Axes`
        an Axes instance to plot to
    opens : sequence
        sequence of opening values
    closes : sequence
        sequence of closing values
    highs : sequence
        sequence of high values
    lows : sequence
        sequence of low values
    width : int
        size of open and close ticks in points
    colorup : color
        the color of the lines where close >= open
    colordown : color
        the color of the lines where close <  open
    alpha : float
        bar transparency

    Returns
    -------
    ret : tuple
        (lineCollection, barCollection)
    """

    return candlestick2_ohlc(ax, opens, highs, lows, closes, width=width,
                             colorup=colorup, colordown=colordown,
                             alpha=alpha)


def candlestick2_ohlc(ax, opens, highs, lows, closes, width=4,
                      colorup='k', colordown='r',
                      alpha=0.75):
    """Represent the open, close as a bar line and high low range as a
    vertical line.

    NOTE: this code assumes if any value open, low, high, close is
    missing they all are missing


    Parameters
    ----------
    ax : `Axes`
        an Axes instance to plot to
    opens : sequence
        sequence of opening values
    highs : sequence
        sequence of high values
    lows : sequence
        sequence of low values
    closes : sequence
        sequence of closing values
    width : int
        size of open and close ticks in points
    colorup : color
        the color of the lines where close >= open
    colordown : color
        the color of the lines where close <  open
    alpha : float
        bar transparency

    Returns
    -------
    ret : tuple
        (lineCollection, barCollection)
    """

    _check_input(opens, highs, lows, closes)

    delta = width / 2.
    barVerts = [((i - delta, open),
                 (i - delta, close),
                 (i + delta, close),
                 (i + delta, open))
                for i, open, close in zip(range(len(opens)), opens, closes)
                if open != -1 and close != -1]

    rangeSegments = [((i, low), (i, high))
                     for i, low, high in zip(range(len(lows)), lows, highs)
                     if low != -1]

    colorup = mcolors.to_rgba(colorup, alpha)
    colordown = mcolors.to_rgba(colordown, alpha)
    colord = {True: colorup, False: colordown}
    colors = [colord[open < close]
              for open, close in zip(opens, closes)
              if open != -1 and close != -1]

    useAA = 0,  # use tuple here
    lw = 0.5,  # and here
    rangeCollection = LineCollection(rangeSegments,
                                     colors=colors,
                                     linewidths=lw,
                                     antialiaseds=useAA,
                                     )

    barCollection = PolyCollection(barVerts,
                                   facecolors=colors,
                                   edgecolors=colors,
                                   antialiaseds=useAA,
                                   linewidths=lw,
                                   )

    minx, maxx = 0, len(rangeSegments)
    miny = min([low for low in lows if low != -1])
    maxy = max([high for high in highs if high != -1])

    corners = (minx, miny), (maxx, maxy)
    ax.update_datalim(corners)
    ax.autoscale_view()

    # add these last
    ax.add_collection(rangeCollection)
    ax.add_collection(barCollection)
    return rangeCollection, barCollection


def volume_overlay(ax, opens, closes, volumes, colorup='g', colordown='r', width=4, alpha=1.0):
    """Add a volume overlay to the current axes.  The opens and closes
    are used to determine the color of the bar.  -1 is missing.  If a
    value is missing on one it must be missing on all

    Parameters
    ----------
    ax : `Axes`
        an Axes instance to plot to
    opens : sequence
        a sequence of opens
    closes : sequence
        a sequence of closes
    volumes : sequence
        a sequence of volumes
    width : int
        the bar width in points
    colorup : color
        the color of the lines where close >= open
    colordown : color
        the color of the lines where close <  open
    alpha : float
        bar transparency

    Returns
    -------
    ret : `barCollection`
        The `barrCollection` added to the axes

    """

    colorup = mcolors.to_rgba(colorup, alpha)
    colordown = mcolors.to_rgba(colordown, alpha)
    colord = {True: colorup, False: colordown}
    colors = [colord[open < close]
              for open, close in zip(opens, closes)
              if open != -1 and close != -1]

    delta = width / 2.
    bars = [((i - delta, 0), (i - delta, v), (i + delta, v), (i + delta, 0))
            for i, v in enumerate(volumes)
            if v != -1]

    barCollection = PolyCollection(bars,
                                   facecolors=colors,
                                   edgecolors=((0, 0, 0, 1),),
                                   antialiaseds=(0,),
                                   linewidths=(0.5,),
                                   )

    ax.add_collection(barCollection)
    corners = (0, 0), (len(bars), max(volumes))
    ax.update_datalim(corners)
    ax.autoscale_view()

    # add these last
    return barCollection


def volume_overlay2(ax, closes, volumes, colorup='g', colordown='r', width=4, alpha=1.0):
    """
    Add a volume overlay to the current axes.  The closes are used to
    determine the color of the bar.  -1 is missing.  If a value is
    missing on one it must be missing on all

    nb: first point is not displayed - it is used only for choosing the
    right color


    Parameters
    ----------
    ax : `Axes`
        an Axes instance to plot to
    closes : sequence
        a sequence of closes
    volumes : sequence
        a sequence of volumes
    width : int
        the bar width in points
    colorup : color
        the color of the lines where close >= open
    colordown : color
        the color of the lines where close <  open
    alpha : float
        bar transparency

    Returns
    -------
    ret : `barCollection`
        The `barrCollection` added to the axes

    """

    return volume_overlay(ax, closes[:-1], closes[1:], volumes[1:], colorup, colordown, width, alpha)


def volume_overlay3(ax, quotes, colorup='g', colordown='r', width=4, alpha=1.0):
    """Add a volume overlay to the current axes.  quotes is a list of (d,
    open, high, low, close, volume) and close-open is used to
    determine the color of the bar

    Parameters
    ----------
    ax : `Axes`
        an Axes instance to plot to
    quotes : sequence of (time, open, high, low, close, ...) sequences
        data to plot.  time must be in float date format - see date2num
    width : int
        the bar width in points
    colorup : color
        the color of the lines where close1 >= close0
    colordown : color
        the color of the lines where close1 <  close0
    alpha : float
         bar transparency

    Returns
    -------
    ret : `barCollection`
        The `barrCollection` added to the axes


    """

    colorup = mcolors.to_rgba(colorup, alpha)
    colordown = mcolors.to_rgba(colordown, alpha)
    colord = {True: colorup, False: colordown}

    dates, opens, highs, lows, closes, volumes = list(zip(*quotes))
    colors = [colord[close1 >= close0]
              for close0, close1 in zip(closes[:-1], closes[1:])
              if close0 != -1 and close1 != -1]
    colors.insert(0, colord[closes[0] >= opens[0]])

    right = width / 2.0
    left = -width / 2.0

    bars = [((left, 0), (left, volume), (right, volume), (right, 0))
            for d, open, high, low, close, volume in quotes]

    sx = ax.figure.dpi * (1.0 / 72.0)  # scale for points
    sy = ax.bbox.height / ax.viewLim.height

    barTransform = Affine2D().scale(sx, sy)

    dates = [d for d, open, high, low, close, volume in quotes]
    offsetsBars = [(d, 0) for d in dates]

    useAA = 0,  # use tuple here
    lw = 0.5,  # and here
    barCollection = PolyCollection(bars,
                                   facecolors=colors,
                                   edgecolors=((0, 0, 0, 1),),
                                   antialiaseds=useAA,
                                   linewidths=lw,
                                   offsets=offsetsBars,
                                   transOffset=ax.transData,
                                   )
    barCollection.set_transform(barTransform)

    minpy, maxx = (min(dates), max(dates))
    miny = 0
    maxy = max([volume for d, open, high, low, close, volume in quotes])
    corners = (minpy, miny), (maxx, maxy)
    ax.update_datalim(corners)
    # print 'datalim', ax.dataLim.bounds
    # print 'viewlim', ax.viewLim.bounds

    ax.add_collection(barCollection)
    ax.autoscale_view()

    return barCollection


def index_bar(ax, vals, facecolor='b', edgecolor='l', width=4, alpha=1.0):
    """Add a bar collection graph with height vals (-1 is missing).

    Parameters
    ----------
    ax : `Axes`
        an Axes instance to plot to
    vals : sequence
        a sequence of values
    facecolor : color
        the color of the bar face
    edgecolor : color
        the color of the bar edges
    width : int
        the bar width in points
    alpha : float
       bar transparency

    Returns
    -------
    ret : `barCollection`
        The `barrCollection` added to the axes

    """

    facecolors = (mcolors.to_rgba(facecolor, alpha),)
    edgecolors = (mcolors.to_rgba(edgecolor, alpha),)

    right = width / 2.0
    left = -width / 2.0

    bars = [((left, 0), (left, v), (right, v), (right, 0))
            for v in vals if v != -1]

    sx = ax.figure.dpi * (1.0 / 72.0)  # scale for points
    sy = ax.bbox.height / ax.viewLim.height

    barTransform = Affine2D().scale(sx, sy)

    offsetsBars = [(i, 0) for i, v in enumerate(vals) if v != -1]

    barCollection = PolyCollection(bars,
                                   facecolors=facecolors,
                                   edgecolors=edgecolors,
                                   antialiaseds=(0,),
                                   linewidths=(0.5,),
                                   offsets=offsetsBars,
                                   transOffset=ax.transData,
                                   )
    barCollection.set_transform(barTransform)

    minpy, maxx = (0, len(offsetsBars))
    miny = 0
    maxy = max([v for v in vals if v != -1])
    corners = (minpy, miny), (maxx, maxy)
    ax.update_datalim(corners)
    ax.autoscale_view()

    # add these last
    ax.add_collection(barCollection)
    return barCollection


def _to_mdate_index(x):
    return date2num(x.index.to_pydatetime()), x


def ohlc_plot(ohlc: pd.DataFrame, width=0, colorup='#209040', colordown='#e02020', fmt=None, autofmt=False):
    """
    Plot OHLC data frame

    :param ohlc: index, 'open', 'high', 'low', 'close' columns
    :param width: used bar width
    :param colorup: color of growing bar
    :param colordown: color of declining bar
    :param autofmt: true if needed to aoutoformatting time labels
    :param fmt: format string for timescale
    :return: axis
    """
    ohlc_f = ohlc.filter(['open', 'high', 'low', 'close'])
    if ohlc_f.shape[1] != 4:
        raise ValueError("DataFrame ohlc must contain 'open', 'high', 'low', 'close' columns !")

    _freq = pd.Timedelta(infer_series_frequency(ohlc_f))

    # customization of the axis
    f = plt.gcf()
    ax = plt.gca()
    ax.xaxis.set_ticks_position('bottom')
    ax.yaxis.set_ticks_position('left')
    ax.tick_params(axis='both', direction='out', width=2, length=8, labelsize=8, pad=8)

    tms, _ = _to_mdate_index(ohlc_f)
    t_data = np.vstack((np.array(tms), ohlc_f.values.T)).T

    # auto width
    if width <= 0:
        width = max(1, _freq.total_seconds() * 0.7) / 24 / 60 / 60

    reshaped_data = np.hstack((np.reshape(t_data[:, 0], (-1, 1)), t_data[:, 1:]))
    candlestick_ohlc(ax, reshaped_data, width=width, colorup=colorup, colordown=colordown)

    is_eod = _freq >= datetime.timedelta(1)
    if is_eod:
        ax.xaxis.set_major_locator(mticker.MaxNLocator(15))
        fmt = '%d-%b-%y' if fmt is None else fmt
    else:
        ax.xaxis.set_major_locator(mticker.MaxNLocator(20))
        fmt = '%H:%M' if fmt is None else fmt

    ticks_loc = ax.get_xticks().tolist()
    ax.xaxis.set_major_locator(mticker.FixedLocator(ticks_loc))
    ax.set_xticklabels([datetime.date.strftime(num2date(x), fmt) for x in ax.get_xticks()])

    if autofmt:
        f.autofmt_xdate()

    # - don't want to see grid lines on the top
    ax.set_axisbelow(True)

    return ax
